"""
    Copyright 2024 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import abc
import base64
import configparser
import dataclasses
import json
import logging
import re
import ssl
import time
from typing import Any, Literal, Mapping, MutableMapping, Optional, Sequence
from urllib import error, request

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers

from inmanta import config, const

from . import exceptions

claim_type = Mapping[str, str | Sequence[str]]


class ClaimMatch:
    """A base class for all claim matching"""

    claim: str
    operator: str
    value: str

    @abc.abstractmethod
    def match_claim(self, claims: claim_type) -> bool:
        """Match the claim

        :param claims: A dict of all claims
        """
        logging.getLogger(__name__).info(f"Matching claims against: '{self}'")
        return False


@dataclasses.dataclass
class InClaim(ClaimMatch):
    """An in claim: exact match of a string in a claim that is a list"""

    claim: str
    value: str
    operator: Literal["in"] = "in"

    def match_claim(self, claims: claim_type) -> bool:
        super().match_claim(claims)

        if self.claim not in claims:
            return False

        claim_value = claims[self.claim]

        if not isinstance(claim_value, list):
            raise ValueError(f"claim {self.claim} should be of type list and not {type(claim_value)}")

        return self.value in claim_value

    def __repr__(self) -> str:
        return f"{self.value} {self.operator} {self.claim}"


@dataclasses.dataclass
class IsClaim(ClaimMatch):
    """An is claim: exact match of a string claim"""

    claim: str
    value: str
    operator: Literal["is"] = "is"

    def match_claim(self, claims: claim_type) -> bool:
        super().match_claim(claims)

        if self.claim not in claims:
            return False

        claim_value = claims[self.claim]

        if not isinstance(claim_value, str):
            raise ValueError(f"claim {self.claim} should be of type str and not {type(claim_value)}")

        return self.value == claim_value

    def __repr__(self) -> str:
        return f"{self.claim} {self.operator} {self.value}"


def check_custom_claims(claims: claim_type, claim_constraints: list[ClaimMatch]) -> bool:
    """Check if the given dict of claims matches the list of constraints. If any of the
    constraints fail, it will return false. If the wrong operation is used on a claim
    it will also result in false. For example, the in operator on a string instead of a
    list of strings

    :param claims: The dict of claims to validate
    :param claim_constraints: A list of all constraints
    :return: The result of the check
    """
    try:
        return all(constraint.match_claim(claims) for constraint in claim_constraints)
    except Exception as e:
        logging.getLogger(__name__).info(
            f"The configured claim constraints failed to evaluate against the provided claims: {e}"
        )
        return False


def encode_token(
    client_types: list[str],
    environment: Optional[str] = None,
    idempotent: bool = False,
    expire: Optional[float] = None,
    custom_claims: Optional[dict[str, str | list[str]]] = None,
) -> str:
    cfg = AuthJWTConfig.get_sign_config()
    if cfg is None:
        raise Exception("No JWT signing configuration available.")

    for ct in client_types:
        if ct not in cfg.client_types:
            raise Exception(
                f"The signing config does not support the requested client type {ct}. Only {cfg.client_types} are allowed."
            )

    payload: dict[str, Any] = {"iss": cfg.issuer, "aud": [cfg.audience], const.INMANTA_URN + "ct": ",".join(client_types)}

    if custom_claims:
        payload.update(custom_claims)

    if not idempotent:
        payload["iat"] = int(time.time())

        if cfg.expire > 0:
            payload["exp"] = int(time.time() + cfg.expire)
        elif expire is not None:
            payload["exp"] = int(time.time() + expire)

    if environment is not None:
        payload[const.INMANTA_URN + "env"] = environment

    return jwt.encode(payload=payload, key=cfg.key, algorithm=cfg.algo)


def decode_token(token: str) -> tuple[claim_type, "AuthJWTConfig"]:
    try:
        # First decode the token without verification
        header = jwt.get_unverified_header(token)
        payload = jwt.decode(token, options={"verify_signature": False})
    except Exception:
        raise exceptions.Forbidden("Unable to decode provided JWT bearer token.")

    if "iss" not in payload:
        raise exceptions.Forbidden("Issuer is required in token to validate.")

    cfg = AuthJWTConfig.get_issuer(str(payload["iss"]))
    if cfg is None:
        raise exceptions.Forbidden("Unknown issuer for token")

    if "alg" not in header or not isinstance(header["alg"], str):
        raise exceptions.Forbidden("alg field is missing in jwt header or is not a valid string")

    alg = header["alg"].lower()
    if alg == "hs256":
        key = cfg.key
    elif alg == "rs256":
        if "kid" not in header or not isinstance(header["kid"], str):
            raise exceptions.Forbidden("kid is missing in jwt header or is not a valid string")
        kid = header["kid"]
        if kid not in cfg.keys:
            raise exceptions.Forbidden(
                "The kid provided in the token does not match a known key. Check the jwks_uri or try "
                "restarting the server to load any new keys."
            )

        key = cfg.keys[kid]
    else:
        raise exceptions.Forbidden("Algorithm %s is not supported." % alg)

    try:
        # copy the payload and make sure the type is claim_type
        decoded_payload: MutableMapping[str, str | Sequence[str]] = {}
        unsupported = []
        for k, v in jwt.decode(token, key, audience=cfg.audience, algorithms=[cfg.algo]).items():
            match v:
                case str():
                    decoded_payload[k] = v
                case list():
                    for el in v:
                        if not isinstance(el, str):
                            raise exceptions.Forbidden(
                                "Only claims of type string or list of strings are supported. "
                                f"Element {el} in claim {k} is not a string."
                            )
                    decoded_payload[k] = v
                case _:
                    unsupported.append(k)

        if unsupported:
            logging.getLogger(__name__).debug(
                "Only claims of type string or list of strings are supported. %s are filtered out.", ", ".join(unsupported)
            )

        ct_key = const.INMANTA_URN + "ct"
        ct_value = str(payload.get(ct_key, "api"))
        decoded_payload[ct_key] = [x.strip() for x in ct_value.split(",")]
    except Exception as e:
        raise exceptions.Forbidden(*e.args)

    if not check_custom_claims(claims=decoded_payload, claim_constraints=cfg.claims):
        raise exceptions.Forbidden("The configured claims constraints did not match. See logs for details.")

    return decoded_payload, cfg


#############################
# auth
#############################
AUTH_JWT_PREFIX = "auth_jwt_"
AUTH_JWT_CLAIM_RE = r"^\s*([^\s].+)\s+(in|is)\s+([^\s].+)\s*$"


class AuthJWTConfig:
    """
    Auth JWT configuration manager
    """

    sections: dict[str, "AuthJWTConfig"] = {}
    issuers: dict[str, "AuthJWTConfig"] = {}

    validate_cert: bool
    key: bytes

    @classmethod
    def reset(cls) -> None:
        cls.sections = {}
        cls.issuers = {}

    @classmethod
    def list(cls) -> list[str]:
        """
        Return a list of all defined auth jwt configurations. This method will load new sections if they were added
        since the last invocation.
        """
        cfg = config.Config.get_instance()
        prefix_len = len(AUTH_JWT_PREFIX)

        for config_section in cfg.keys():
            if config_section[:prefix_len] == AUTH_JWT_PREFIX:
                name = config_section[prefix_len:]
                if name not in cls.sections:
                    obj = cls(name, config_section, cfg[config_section])
                    cls.sections[name] = obj
                    if obj.issuer in cls.issuers:
                        raise ValueError("Only one configuration per issuer is supported")

                    cls.issuers[obj.issuer] = obj

        # Verify that only one has sign set to true
        sign = False
        for section in cls.sections.values():
            if section.sign:
                if sign:
                    raise ValueError("Only one auth_jwt section may have sign set to true")
                else:
                    sign = True

        if len(cls.sections.keys()) > 0 and not sign:
            raise ValueError("One auth_jwt section should have sign set to true")

        return list(cls.sections.keys())

    @classmethod
    def get(cls, name: str) -> Optional["AuthJWTConfig"]:
        """
        Get the config with the given name
        """
        cls.list()
        if name in cls.sections:
            return cls.sections[name]
        return None

    @classmethod
    def get_sign_config(cls) -> Optional["AuthJWTConfig"]:
        """
        Get the configuration with sign is true
        """
        cls.list()
        for cfg in cls.sections.values():
            if cfg.sign:
                return cfg
        return None

    @classmethod
    def get_issuer(cls, issuer: str) -> Optional["AuthJWTConfig"]:
        """
        Get the config for the given issuer. Only when no auth config has been loaded yet, the configuration will be loaded
        again. For loading additional configuration, call list() first. This method is in the auth path for each API
        request.
        """
        if len(cls.issuers) == 0:
            cls.list()
        if issuer in cls.issuers:
            return cls.issuers[issuer]
        return None

    def __init__(self, name: str, section: str, config: configparser.SectionProxy) -> None:
        self.name: str = name
        self.section: str = section
        self.keys: dict[str, bytes] = {}
        self._config: configparser.SectionProxy = config
        self.claims: list[ClaimMatch] = []

        self.jwt_username_claim: str = "sub"
        self.expire: int = 0
        self.sign: bool = False
        self.issuer: str = "https://localhost:8888/"
        self.audience: str

        if "algorithm" not in config:
            raise ValueError("algorithm is required in %s section" % self.section)

        self.algo = config["algorithm"]
        self.validate_generic()

        if self.algo.lower() == "hs256":
            self.validate_hs265()
        elif self.algo.lower() == "rs256":
            self.validate_rs265()
        else:
            raise ValueError(f"Algorithm {self.algo} in {self.section} is not support ")

    def validate_generic(self) -> None:
        """
        Validate  and parse the generic options that are valid for all algorithms
        """
        if "sign" in self._config:
            self.sign = config.is_bool(self._config["sign"])

        if "client_types" not in self._config:
            raise ValueError("client_types is a required options for %s" % self.section)

        self.client_types = config.is_list(self._config["client_types"])
        for ct in self.client_types:
            if ct not in [client_type for client_type in const.ClientType]:
                raise ValueError(f"invalid client_type {ct} in {self.section}")

        if "expire" in self._config:
            self.expire = config.is_int(self._config["expire"])

        if "issuer" in self._config:
            self.issuer = config.is_str(self._config["issuer"])

        if "audience" in self._config:
            self.audience = config.is_str(self._config["audience"])
        else:
            self.audience = self.issuer

        if "claims" in self._config:
            self.parse_claim_matching(self._config["claims"])

        if "jwt-username-claim" in self._config:
            if self.sign:
                raise ValueError(f"auth config {self.section} used for signing cannot use a custom claim.")
            self.jwt_username_claim = self._config.get("jwt-username-claim")

    def parse_claim_matching(self, claim_conf: str) -> None:
        """Parse claim matching expressions"""
        items = re.findall(AUTH_JWT_CLAIM_RE, claim_conf, re.MULTILINE)
        for item in items:
            match item:
                case (claim, "is", value):
                    self.claims.append(IsClaim(claim, value))
                case (value, "in", claim):
                    self.claims.append(InClaim(claim, value))
                case _:
                    raise ValueError(f"Invalid claim match '{' '.join(item)}' in {self.section}")

    def validate_hs265(self) -> None:
        """
        Validate and parse HS256 algorithm configuration
        """
        if "key" not in self._config:
            raise ValueError(f"key is required in {self.section} for algorithm {self.algo}")

        self.key = base64.urlsafe_b64decode((self._config["key"] + "==").encode("ascii"))
        if len(self.key) < 32:
            raise ValueError("HS256 requires a key of 32 bytes (256 bits) or longer in " + self.section)

    def _load_public_key(self, e: str, n: str) -> bytes:
        def to_int(x: str) -> int:
            bs = base64.urlsafe_b64decode(x + "==")
            return int.from_bytes(bs, byteorder="big")

        ei = to_int(e)
        ni = to_int(n)
        numbers = RSAPublicNumbers(ei, ni)
        public_key = numbers.public_key(backend=default_backend())
        pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        return pem

    def validate_rs265(self) -> None:
        """
        Validate and parse RS256 algorithm configuration
        """
        if "jwks_uri" not in self._config:
            raise ValueError("jwks_uri is required for RS256 based providers in section %s" % self.section)

        self.jwks_uri = self._config["jwks_uri"]

        if "validate_cert" in self._config:
            self.validate_cert = self._config.getboolean("validate_cert")
        else:
            self.validate_cert = True

        ctx = None
        if not self.validate_cert:
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        jwks_timeout = self._config.getfloat("jwks_request_timeout", 30.0)
        try:
            with request.urlopen(self.jwks_uri, timeout=jwks_timeout, context=ctx) as response:
                key_data = json.loads(response.read().decode("utf-8"))
        except error.URLError as e:
            # HTTPError is raised for non-200 responses; the response
            # can be found in e.response.
            raise ValueError(
                "Unable to load key data for %s using the provided jwks_uri %s. Got error: %s"
                % (self.section, self.jwks_uri, e.reason)
            )
        except Exception:
            # Other errors are possible, such as IOError.
            raise ValueError("Unable to load key data for %s using the provided jwks_uri." % self.section)

        for key in key_data["keys"]:
            self.keys[key["kid"]] = self._load_public_key(key["e"], key["n"])



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import asyncio
import inspect
import logging
import socket
import uuid
from asyncio import CancelledError, run_coroutine_threadsafe, sleep
from collections import abc, defaultdict
from collections.abc import Coroutine
from enum import Enum
from typing import Any, Callable, Optional
from urllib import parse

import pydantic

from inmanta import config as inmanta_config
from inmanta import types, util
from inmanta.protocol import common, exceptions
from inmanta.util import TaskHandler

from .rest import client

LOGGER: logging.Logger = logging.getLogger(__name__)


class CallTarget:
    """
    A baseclass for all classes that are target for protocol calls / methods
    """

    def _get_endpoint_metadata(self) -> dict[str, list[tuple[str, Callable]]]:
        total_dict = {
            method_name: method
            for method_name, method in inspect.getmembers(self)
            if callable(method) and method_name[0] != "_"
        }

        methods: dict[str, list[tuple[str, Callable]]] = defaultdict(list)
        for name, attr in total_dict.items():
            if hasattr(attr, "__protocol_method__"):
                methods[attr.__protocol_method__.__name__].append((name, attr))

        return methods

    def get_op_mapping(self) -> dict[str, dict[str, common.UrlMethod]]:
        """
        Build a mapping between urls, ops and methods
        """
        url_map: dict[str, dict[str, common.UrlMethod]] = defaultdict(dict)

        # Loop over all methods in this class that have a handler annotation. The handler annotation refers to a method
        # definition. This method definition defines how the handler is invoked.
        for method, handler_list in self._get_endpoint_metadata().items():
            for method_handlers in handler_list:
                # Go over all method annotation on the method associated with the handler
                for properties in common.MethodProperties.methods[method]:
                    url = properties.get_listen_url()

                    # Associate the method with the handler if:
                    # - the handler does not specific a method version
                    # - the handler specifies a method version and the method version matches the method properties
                    if method_handlers[1].__api_version__ is None or (
                        method_handlers[1].__api_version__ is not None
                        and properties.api_version == method_handlers[1].__api_version__
                    ):
                        # there can only be one
                        if url in url_map and properties.operation in url_map[url]:
                            raise Exception(f"A handler is already registered for {properties.operation} {url}. ")

                        url_map[url][properties.operation] = common.UrlMethod(
                            properties, self, method_handlers[1], method_handlers[0]
                        )
        return url_map


class Endpoint(TaskHandler[None]):
    """
    An end-point in the rpc framework
    """

    def __init__(self, name: str):
        super().__init__()
        self._name: str = name
        self._node_name: str = inmanta_config.nodename.get()
        self._end_point_names: set[str] = set()
        self._targets: list[CallTarget] = []

    def add_call_target(self, target: CallTarget) -> None:
        self._targets.append(target)

    @property
    def call_targets(self) -> list[CallTarget]:
        return self._targets

    def get_end_point_names(self) -> set[str]:
        return self._end_point_names

    async def add_end_point_name(self, name: str) -> None:
        """
        Add an additional name to this endpoint to which it reacts and sends out in heartbeats
        """
        LOGGER.debug("Adding '%s' as endpoint", name)
        self._end_point_names.add(name)

    async def remove_end_point_name(self, name: str) -> None:
        LOGGER.debug("Removing '%s' as endpoint", name)
        self._end_point_names.discard(name)

    def clear_end_points(self) -> None:
        """
        Clear all endpoints
        """
        self._end_point_names = set()

    name = property(lambda self: self._name)
    end_point_names = property(get_end_point_names)

    def _get_hostname(self) -> str:
        """
        Determine the hostname of this machine
        """
        return socket.gethostname()

    def get_node_name(self) -> str:
        return self._node_name

    node_name = property(get_node_name)

    async def stop(self) -> None:
        """Stop this endpoint"""
        await super().stop()


class SessionEndpoint(Endpoint, CallTarget):
    """
    An endpoint for clients that make calls to a server and that receive calls back from the server using long-poll
    """

    _client: "SessionClient"
    _heartbeat_client: "SessionClient"

    def __init__(self, name: str, timeout: int = 120, reconnect_delay: int = 5):
        super().__init__(name)
        self._transport = client.RESTClient
        self._sched = util.Scheduler("session endpoint")

        self._env_id: Optional[uuid.UUID] = None

        self.sessionid: uuid.UUID = uuid.uuid1()
        self.running: bool = True
        self.server_timeout = timeout
        self.reconnect_delay = reconnect_delay
        self.add_call_target(self)

        self._client = SessionClient(self.name, self.sessionid, timeout=self.server_timeout)
        self._heartbeat_client = SessionClient(self.name, self.sessionid, timeout=self.server_timeout, force_instance=True)

    def get_environment(self) -> Optional[uuid.UUID]:
        return self._env_id

    @property
    def environment(self) -> Optional[uuid.UUID]:
        return self._env_id

    def set_environment(self, environment_id: uuid.UUID) -> None:
        """
        Set the environment of this agent
        """
        if isinstance(environment_id, str):
            self._env_id = uuid.UUID(environment_id)
        else:
            self._env_id = environment_id

    async def start_connected(self) -> None:
        """
        This method is called after starting the client transport, but before sending the first heartbeat.
        """

    async def start(self) -> None:
        """
        Connect to the server and use a heartbeat and long-poll for two-way communication
        """
        assert self._env_id is not None
        LOGGER.info("Starting agent for %s", str(self.sessionid))
        await self.start_connected()
        self.add_background_task(self.perform_heartbeat())

    async def stop(self) -> None:
        self._heartbeat_client.close()
        await self._sched.stop()
        await super().stop()

    async def on_reconnect(self) -> None:
        """
        Called when a connection becomes active. i.e. when a first heartbeat is received after startup or
        a first hearbeat after an :py:`on_disconnect`
        """

    async def on_disconnect(self) -> None:
        """
        Called when the connection is lost unexpectedly (not on shutdown)
        """

    async def perform_heartbeat(self) -> None:
        """
        Start a continuous heartbeat call
        """
        if self._heartbeat_client is None or self._client is None:
            raise Exception("AgentEndpoint not started")

        connected: bool = False
        try:
            while True:
                LOGGER.log(3, "sending heartbeat for %s", str(self.sessionid))
                result = await self._heartbeat_client.heartbeat(
                    sid=str(self.sessionid),
                    tid=str(self._env_id),
                    endpoint_names=list(self.end_point_names),
                    nodename=self.node_name,
                    no_hang=not connected,
                )
                LOGGER.log(3, "returned heartbeat for %s", str(self.sessionid))
                if result.code == 200:
                    if not connected:
                        connected = True
                        self.add_background_task(self.on_reconnect())
                    if result.result is not None:
                        if "method_calls" in result.result:
                            method_calls: list[common.Request] = [
                                common.Request.from_dict(req) for req in result.result["method_calls"]
                            ]
                            # FIXME: reuse transport?
                            transport = self._transport(self)

                            for method_call in method_calls:
                                self.add_background_task(self.dispatch_method(transport, method_call))
                else:
                    LOGGER.warning(
                        "Heartbeat failed with status %d and message: %s, going to sleep for %d s",
                        result.code,
                        result.result,
                        self.reconnect_delay,
                    )
                    connected = False
                    await self.on_disconnect()
                    await sleep(self.reconnect_delay)
        except CancelledError:
            pass

    async def dispatch_method(self, transport: client.RESTClient, method_call: common.Request) -> None:
        if self._client is None:
            raise Exception("AgentEndpoint not started")

        LOGGER.debug("Received call through heartbeat: %s %s %s", method_call.reply_id, method_call.method, method_call.url)
        kwargs, config = transport.match_call(method_call.url, method_call.method)

        if config is None:
            msg = "An error occurred during heartbeat method call ({} {} {}): {}".format(
                method_call.reply_id,
                method_call.method,
                method_call.url,
                "No such method",
            )
            LOGGER.error(msg)
            # if reply_id is none, we don't send the reply
            if method_call.reply_id is not None:
                await self._client.heartbeat_reply(self.sessionid, method_call.reply_id, {"result": msg, "code": 500})
            return

        body = method_call.body or {}
        query_string = parse.urlparse(method_call.url).query
        for key, value in parse.parse_qs(query_string, keep_blank_values=True):
            if len(value) == 1:
                body[key] = value[0].decode("latin-1")
            else:
                body[key] = [v.decode("latin-1") for v in value]

        body.update(kwargs)

        response: common.Response = await transport._execute_call(config, body, method_call.headers)

        if response.status_code == 500:
            msg = ""
            if response.body is not None and "message" in response.body:
                msg = response.body["message"]
            LOGGER.error(
                "An error occurred during heartbeat method call (%s %s %s): %s",
                method_call.reply_id,
                method_call.method,
                method_call.url,
                msg,
            )

        if self._client is None:
            raise Exception("AgentEndpoint not started")

        # if reply is is none, we don't send the reply
        if method_call.reply_id is not None:
            await self._client.heartbeat_reply(
                self.sessionid, method_call.reply_id, {"result": response.body, "code": response.status_code}
            )


class VersionMatch(str, Enum):
    lowest = "lowest"
    """ Select the lowest available version of the method
    """
    highest = "highest"
    """ Select the highest available version of the method
    """
    exact = "exact"
    """ Select the exact version of the method
    """


class Client(Endpoint):
    """
    A client that communicates with end-point based on its configuration
    """

    def __init__(
        self,
        name: str,
        timeout: int = 120,
        version_match: VersionMatch = VersionMatch.lowest,
        exact_version: int = 0,
        with_rest_client: bool = True,
        force_instance: bool = False,
    ) -> None:
        super().__init__(name)
        assert isinstance(timeout, int), "Timeout needs to be an integer value."
        LOGGER.debug("Start transport for client %s", self.name)
        if with_rest_client:
            self._transport_instance = client.RESTClient(self, connection_timout=timeout, force_instance=force_instance)
        else:
            self._transport_instance = None
        self._version_match = version_match
        self._exact_version = exact_version

    def close(self):
        """
        Closes the RESTclient instance manually. This is only needed when it is started with force_instance set to true
        """
        self._transport_instance.close()

    async def _call(
        self, method_properties: common.MethodProperties, args: list[object], kwargs: dict[str, object]
    ) -> common.Result:
        """
        Execute a call and return the result
        """
        result = await self._transport_instance.call(method_properties, args, kwargs)
        return result

    def _select_method(self, name) -> Optional[common.MethodProperties]:
        if name not in common.MethodProperties.methods:
            return None

        methods = common.MethodProperties.methods[name]

        if self._version_match is VersionMatch.lowest:
            return min(methods, key=lambda x: x.api_version)
        elif self._version_match is VersionMatch.highest:
            return max(methods, key=lambda x: x.api_version)
        elif self._version_match is VersionMatch.exact:
            return next((m for m in methods if m.api_version == self._exact_version), None)

        return None

    def __getattr__(self, name: str) -> Callable[..., Coroutine[Any, Any, common.Result]]:
        """
        Return a function that will call self._call with the correct method properties associated
        """
        method = self._select_method(name)

        if method is None:
            raise AttributeError("Method with name %s is not defined for this client" % name)

        def wrap(*args: object, **kwargs: object) -> Coroutine[Any, Any, common.Result]:
            assert method
            method.function(*args, **kwargs)
            return self._call(method_properties=method, args=args, kwargs=kwargs)

        return wrap


class SyncClient:
    """
    A synchronous client that communicates with end-point based on its configuration
    """

    def __init__(
        self,
        name: Optional[str] = None,
        timeout: int = 120,
        client: Optional[Client] = None,
        ioloop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        """
        either name or client is required.
        they can not be used at the same time

        :param name: name of the configuration to use for this endpoint. The config section used is "{name}_rest_transport"
        :param client: the client to use for this sync_client
        :param timeout: http timeout on all requests

        :param ioloop: the specific (running) ioloop to schedule this request on. The loop should run on a different thread
            than the one the client methods are called on. If no ioloop is passed, we assume there is no running ioloop in the
            context where this syncclient is used.
        """
        if (name is None) == (client is None):
            # Exactly one must be set
            raise Exception("Either name or client needs to be provided.")

        self.timeout = timeout
        self._ioloop: Optional[asyncio.AbstractEventLoop] = ioloop
        if client is None:
            assert name is not None  # Make mypy happy
            self.name = name
            self._client = Client(name, self.timeout)
        else:
            self.name = client.name
            self._client = client

    def __getattr__(self, name: str) -> Callable[..., common.Result]:
        def async_call(*args: list[object], **kwargs: dict[str, object]) -> common.Result:
            method: Callable[..., abc.Awaitable[common.Result]] = getattr(self._client, name)
            with_timeout: abc.Awaitable[common.Result] = asyncio.wait_for(method(*args, **kwargs), self.timeout)

            try:
                if self._ioloop is None:
                    # no loop is running: create a loop for this thread if it doesn't exist already and run it
                    return util.ensure_event_loop().run_until_complete(with_timeout)
                else:
                    # loop is running on different thread
                    return run_coroutine_threadsafe(with_timeout, self._ioloop).result()
            except TimeoutError:
                raise ConnectionRefusedError()

        return async_call


class SessionClient(Client):
    """
    A client that communicates with server endpoints over a session.
    """

    def __init__(self, name: str, sid: uuid.UUID, timeout: int = 120, force_instance: bool = False) -> None:
        super().__init__(name, timeout, force_instance=force_instance)
        self._sid = sid

    async def _call(
        self, method_properties: common.MethodProperties, args: list[object], kwargs: dict[str, object]
    ) -> common.Result:
        """
        Execute the rpc call
        """
        if "sid" not in kwargs:
            kwargs["sid"] = self._sid

        result = await self._transport_instance.call(method_properties, args, kwargs)
        return result


class TypedClient(Client):
    """A client that returns typed data instead of JSON"""

    def _raise_exception(self, exception_class: type[exceptions.BaseHttpException], result: Optional[types.JsonType]) -> None:
        """Raise an exception based on the provided status"""
        if result is None:
            raise exception_class()

        message = result.get("message", None)
        details = result.get("error_details", None)

        raise exception_class(message, details)

    def _process_response(self, method_properties: common.MethodProperties, response: common.Result) -> types.ReturnTypes:
        """Convert the response into a proper type and restore exception if any"""
        match response.code:
            case 200:
                # typed methods always require an envelope key
                if response.result is None or method_properties.envelope_key not in response.result:
                    raise exceptions.BadRequest("No data was provided in the body. Make sure to only use typed methods.")

                if method_properties.return_type is None:
                    return None

                try:
                    ta = pydantic.TypeAdapter(method_properties.return_type)
                except common.InvalidMethodDefinition:
                    raise exceptions.BadRequest("Typed client can only be used with typed methods.")

                return ta.validate_python(response.result[method_properties.envelope_key])

            case 400:
                self._raise_exception(exceptions.BadRequest, response.result)

            case 401:
                self._raise_exception(exceptions.UnauthorizedException, response.result)

            case 403:
                self._raise_exception(exceptions.Forbidden, response.result)

            case 404:
                self._raise_exception(exceptions.NotFound, response.result)

            case 409:
                self._raise_exception(exceptions.Conflict, response.result)

            case 500:
                self._raise_exception(exceptions.ServerError, response.result)

            case 503:
                self._raise_exception(exceptions.ShutdownInProgress, response.result)

            case _:
                self._raise_exception(exceptions.ServerError, response.result)

        # make mypy happy, it cannot deduce that all the cases will always raise an exception
        return None

    async def _call(
        self, method_properties: common.MethodProperties, args: list[object], kwargs: dict[str, object]
    ) -> types.ReturnTypes:
        """Execute a call and return the result"""
        return self._process_response(method_properties, await super()._call(method_properties, args, kwargs))



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import inspect
from typing import Callable, Optional, TypeVar, Union

from inmanta import const
from inmanta.types import Apireturn, HandlerType, MethodType

from . import common

FuncT = TypeVar("FuncT", bound=HandlerType)


class handle:  # noqa: N801
    """
    Decorator for subclasses of an endpoint to handle protocol methods

    :param method: A subclass of method that defines the method
    :param api_version: When specific this handler is only associated with a method of the specific api version. If the
                        version is not defined, the handler is not associated with a rest endpoint.
    :param kwargs: Map arguments in the message from one name to an other
    """

    def __init__(self, method: Callable[..., Apireturn], api_version: Optional[int] = None, **kwargs: str) -> None:
        self.method = method
        self.mapping: dict[str, str] = kwargs
        self._api_version = api_version

    def __call__(self, function: FuncT) -> FuncT:
        """
        The wrapping
        """
        if not inspect.iscoroutinefunction(function):
            raise ValueError(f"{function} is not an async function. Only async def functions may handle requests.")

        function.__protocol_method__ = self.method
        function.__protocol_mapping__ = self.mapping
        function.__api_version__ = self._api_version
        return function


MethodT = TypeVar("MethodT", bound=MethodType)


def method(
    path: str,
    operation: str = "POST",
    reply: bool = True,
    arg_options: dict[str, common.ArgOption] = {},
    timeout: Optional[int] = None,
    server_agent: bool = False,
    api: Optional[bool] = None,
    agent_server: bool = False,
    validate_sid: Optional[bool] = None,
    client_types: list[const.ClientType] = [const.ClientType.api],
    api_version: int = 1,
    api_prefix: str = "api",
    envelope: bool = False,
    envelope_key: str = const.ENVELOPE_KEY,
    enforce_auth: bool = True,
) -> Callable[..., Callable]:
    """
    Decorator to identify a method as a RPC call. The arguments of the decorator are used by each transport to build
    and model the protocol.

    :param path: The url path to use for this call. This path can contain parameter names of the function. These names
        should be enclosed in < > brackets.
    :param operation: The type of HTTP operation (verb).
    :param timeout: nr of seconds before request it terminated.
    :param api: This is a call from the client to the Server (True if not server_agent and not agent_server).
    :param server_agent: This is a call from the Server to the Agent (reverse http channel through long poll).
    :param agent_server: This is a call from the Agent to the Server.
    :param validate_sid: This call requires a valid session, true by default if agent_server and not api
    :param client_types: The allowed client types for this call.
        The valid values are defined by the :const:`inmanta.const.ClientType` enum.
    :param arg_options: Options related to arguments passed to the method. The key of this dict is the name of the arg to
        which the options apply. The value is another dict that can contain the following options:

            header: Map this argument to a header with the following name.
            reply_header: If the argument is mapped to a header, this header will also be included in the reply
            getter: Call this method after validation and pass its return value to the method call. This may change the
            type of the argument. This method can raise an HTTPException to return a 404 for example.

    :param api_version: The version of the api this method belongs to.
    :param api_prefix: The prefix of the method: /<prefix>/v<version>/<method_name>.
    :param envelope: Put the response of the call under an envelope with key envelope_key.
    :param envelope_key: The envelope key to use.
    :param enforce_auth: When set to true authentication is enforced on this endpoint. When set to false, authentication is not
                         enforced, even if auth is enabled.
    """

    def wrapper(func: MethodT) -> MethodT:
        properties = common.MethodProperties(
            func,
            path,
            operation,
            reply,
            arg_options,
            timeout,
            server_agent,
            api,
            agent_server,
            validate_sid,
            client_types,
            api_version,
            api_prefix,
            envelope,
            False,
            envelope_key,
            enforce_auth=enforce_auth,
        )
        common.MethodProperties.register_method(properties)
        return func

    return wrapper


def typedmethod(
    path: Union[str, list[str]],
    operation: str = "POST",
    reply: bool = True,
    arg_options: dict[str, common.ArgOption] = {},
    timeout: Optional[int] = None,
    server_agent: bool = False,
    api: Optional[bool] = None,
    agent_server: bool = False,
    validate_sid: Optional[bool] = None,
    client_types: list[const.ClientType] = [const.ClientType.api],
    api_version: int = 1,
    api_prefix: str = "api",
    envelope_key: str = const.ENVELOPE_KEY,
    strict_typing: bool = True,
    enforce_auth: bool = True,
    varkw: bool = False,
) -> Callable[..., Callable]:
    """
    Decorator to identify a method as a RPC call. The arguments of the decorator are used by each transport to build
    and model the protocol.

    :param path: The url path(s) to use for this call. This path can contain parameter names of the function. These names
                 should be enclosed in < > brackets.
    :param operation: The type of HTTP operation (verb)
    :param timeout: nr of seconds before request it terminated
    :param api: This is a call from the client to the Server (True if not server_agent and not agent_server)
    :param server_agent: This is a call from the Server to the Agent (reverse http channel through long poll)
    :param agent_server: This is a call from the Agent to the Server
    :param validate_sid: This call requires a valid session, true by default if agent_server and not api
    :param client_types: The allowed client types for this call.
            The valid values are defined by the :const:`inmanta.const.ClientType` enum.
    :param arg_options: Options related to arguments passed to the method. The key of this dict is the name of the arg to
        which the options apply. The value is another dict that can contain the following options:
            header: Map this argument to a header with the following name.
            reply_header: If the argument is mapped to a header, this header will also be included in the reply
            getter: Call this method after validation and pass its return value to the method call. This may change the
                    type of the argument. This method can raise an HTTPException to return a 404 for example.
    :param api_version: The version of the api this method belongs to
    :param api_prefix: The prefix of the method: /<prefix>/v<version>/<method_name>
    :param envelope_key: The envelope key to use.
    :param strict_typing: If true, does not allow `Any`. Setting this option to False is heavily discouraged except for some
        few very specific cases where the type system does not allow the strict type to be specified, for example in case of
        infinite recursion.
    :param enforce_auth: When set to true authentication is enforced on this endpoint. When set to false, authentication is not
                         enforced, even if auth is enabled.
    :param varkw: If true, additional arguments are allowed and will be dispatched to the handler. The handler is
                  responsible for the validation.
    """

    def wrapper(func: MethodT) -> MethodT:
        paths = path if isinstance(path, list) else [path]
        for current_path in paths:
            properties = common.MethodProperties(
                func,
                current_path,
                operation,
                reply,
                arg_options,
                timeout,
                server_agent,
                api,
                agent_server,
                validate_sid,
                client_types,
                api_version,
                api_prefix,
                True,
                True,
                envelope_key,
                strict_typing=strict_typing,
                enforce_auth=enforce_auth,
                varkw=varkw,
            )
            common.MethodProperties.register_method(properties)
        return func

    return wrapper



"""
    Copyright 2024 Inmanta
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
        http://www.apache.org/licenses/LICENSE-2.0
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    Contact: code@inmanta.com
"""

import asyncio
import datetime
import logging
import typing
import uuid
from collections.abc import Sequence
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any, Optional

import inmanta.agent.cache
import inmanta.protocol
import inmanta.util
from inmanta import const, data
from inmanta.agent import executor, handler
from inmanta.agent.executor import FailedResourcesSet, ResourceDetails
from inmanta.agent.handler import HandlerAPI, SkipResource
from inmanta.agent.io.remote import ChannelClosedException
from inmanta.const import ParameterSource
from inmanta.data.model import AttributeStateChange, ResourceIdStr, ResourceVersionIdStr
from inmanta.resources import Id, Resource
from inmanta.types import Apireturn

if typing.TYPE_CHECKING:
    import inmanta.agent.agent as agent


class InProcessExecutor(executor.Executor, executor.AgentInstance):
    """
    This is an executor that executes in the process it is started in
    """

    def __init__(
        self,
        agent_name: str,
        agent_uri: str,
        environment: uuid.UUID,
        client: inmanta.protocol.SessionClient,
        eventloop: asyncio.AbstractEventLoop,
        parent_logger: logging.Logger,
    ):
        self.name = agent_name
        self.client = client
        self.uri = agent_uri

        # For agentinstance api
        self.eventloop = eventloop
        self.sessionid = client._sid
        self.environment = environment

        # threads to setup _io ssh connections
        self.provider_thread_pool: ThreadPoolExecutor = ThreadPoolExecutor(1, thread_name_prefix="ProviderPool_%s" % self.name)
        # threads to work
        self.thread_pool: ThreadPoolExecutor = ThreadPoolExecutor(1, thread_name_prefix="Pool_%s" % self.name)

        self._cache = inmanta.agent.cache.AgentCache(self)

        self.logger: logging.Logger = parent_logger.getChild(self.name)

        self._stopped = False

        self.failed_resource_types: FailedResourcesSet = set()

    def stop(self) -> None:
        self._stopped = True
        self._cache.close()
        self.provider_thread_pool.shutdown(wait=False)
        self.thread_pool.shutdown(wait=False)

    def join(self, thread_pool_finalizer: list[ThreadPoolExecutor]) -> None:
        """
        Called after stop to ensure complete shutdown

        :param thread_pool_finalizer: all threadpools that should be joined should be added here.
        """
        assert self._stopped
        thread_pool_finalizer.append(self.provider_thread_pool)
        thread_pool_finalizer.append(self.thread_pool)

    def is_stopped(self) -> bool:
        return self._stopped

    async def get_provider(self, resource: Resource) -> HandlerAPI[Any]:
        provider = await asyncio.get_running_loop().run_in_executor(
            self.provider_thread_pool, handler.Commander.get_provider, self._cache, self, resource
        )
        provider.set_cache(self._cache)
        return provider

    async def send_in_progress(
        self, action_id: uuid.UUID, env_id: uuid.UUID, resource_id: ResourceVersionIdStr
    ) -> dict[ResourceIdStr, const.ResourceState]:
        result = await self.client.resource_deploy_start(
            tid=env_id,
            rvid=resource_id,
            action_id=action_id,
        )
        if result.code != 200 or result.result is None:
            raise Exception("Failed to report the start of the deployment to the server")
        return {Id.parse_id(key).resource_str(): const.ResourceState[value] for key, value in result.result["data"].items()}

    async def _execute(
        self,
        resource: Resource,
        gid: uuid.UUID,
        ctx: handler.HandlerContext,
        requires: dict[ResourceIdStr, const.ResourceState],
    ) -> None:
        """
        Get the handler for a given resource and run its ``deploy`` method.

        :param resource: The resource to deploy.
        :param gid: Id of this deploy.
        :param ctx: The context to use during execution of this deploy.
        :param requires: A dictionary that maps each dependency of the resource to be deployed, to its latest resource
                         state that was not `deploying'.
        """
        # setup provider
        ctx.debug("Start deploy %(deploy_id)s of resource %(resource_id)s", deploy_id=gid, resource_id=resource.id)

        provider: Optional[HandlerAPI[Any]] = None
        try:
            provider = await self.get_provider(resource)
        except ChannelClosedException as e:
            ctx.set_status(const.ResourceState.unavailable)
            ctx.exception(str(e))
            return
        except Exception:
            ctx.set_status(const.ResourceState.unavailable)
            ctx.exception("Unable to find a handler for %(resource_id)s", resource_id=resource.id.resource_version_str())
            return
        else:
            # main execution
            try:
                await asyncio.get_running_loop().run_in_executor(
                    self.thread_pool,
                    provider.deploy,
                    ctx,
                    resource,
                    requires,
                )
                if ctx.status is None:
                    ctx.set_status(const.ResourceState.deployed)
            except ChannelClosedException as e:
                ctx.set_status(const.ResourceState.failed)
                ctx.exception(str(e))
            except SkipResource as e:
                ctx.set_status(const.ResourceState.skipped)
                ctx.warning(msg="Resource %(resource_id)s was skipped: %(reason)s", resource_id=resource.id, reason=e.args)
            except Exception as e:
                ctx.set_status(const.ResourceState.failed)
                ctx.exception(
                    "An error occurred during deployment of %(resource_id)s (exception: %(exception)s",
                    resource_id=resource.id,
                    exception=repr(e),
                )
        finally:
            if provider is not None:
                provider.close()

    async def _report_resource_deploy_done(
        self,
        resource_details: ResourceDetails,
        ctx: handler.HandlerContext,
    ) -> None:
        """
        Report the end of a resource deployment to the server.

        :param resource_details: Deployed resource being reported.
        :param ctx: Context of the associated handler.
        """

        changes: dict[ResourceVersionIdStr, dict[str, AttributeStateChange]] = {resource_details.rvid: ctx.changes}
        response = await self.client.resource_deploy_done(
            tid=resource_details.env_id,
            rvid=resource_details.rvid,
            action_id=ctx.action_id,
            status=ctx.status,
            messages=ctx.logs,
            changes=changes,
            change=ctx.change,
        )
        if response.code != 200:
            self.logger.error("Resource status update failed %s for %s ", response.result, resource_details.rvid)

    async def deserialize(self, resource_details: ResourceDetails, action: const.ResourceAction) -> Optional[Resource]:
        started: datetime.datetime = datetime.datetime.now().astimezone()
        try:
            return Resource.deserialize(resource_details.attributes)
        except Exception:
            msg = data.LogLine.log(
                level=const.LogLevel.ERROR,
                msg="Unable to deserialize %(resource_id)s",
                resource_id=resource_details.rvid,
                timestamp=datetime.datetime.now().astimezone(),
            )

            await self.client.resource_action_update(
                tid=resource_details.env_id,
                resource_ids=[resource_details.rvid],
                action_id=uuid.uuid4(),
                action=action,
                started=started,
                finished=datetime.datetime.now().astimezone(),
                status=const.ResourceState.unavailable,
                messages=[msg],
            )
            raise

    async def execute(
        self,
        gid: uuid.UUID,
        resource_details: ResourceDetails,
        reason: str,
    ) -> None:
        try:
            resource: Resource | None = await self.deserialize(resource_details, const.ResourceAction.deploy)
        except Exception:
            return
        assert resource is not None
        ctx = handler.HandlerContext(resource, logger=self.logger)
        ctx.debug(
            "Start run for resource %(resource)s because %(reason)s",
            resource=str(resource_details.rvid),
            deploy_id=gid,
            agent=self.name,
            reason=reason,
        )

        try:
            requires: dict[ResourceIdStr, const.ResourceState] = await self.send_in_progress(
                ctx.action_id, resource_details.env_id, resource_details.rvid
            )
        except Exception:
            ctx.set_status(const.ResourceState.failed)
            ctx.exception("Failed to report the start of the deployment to the server")
            return

        await self._execute(resource, gid=gid, ctx=ctx, requires=requires)

        ctx.debug(
            "End run for resource %(r_id)s in deploy %(deploy_id)s",
            r_id=resource_details.rvid,
            deploy_id=gid,
        )

        if ctx.facts:
            ctx.debug("Sending facts to the server")
            set_fact_response = await self.client.set_parameters(tid=resource_details.env_id, parameters=ctx.facts)
            if set_fact_response.code != 200:
                ctx.error("Failed to send facts to the server %s", set_fact_response.result)

        await self._report_resource_deploy_done(resource_details, ctx)

    async def dry_run(
        self,
        resources: Sequence[ResourceDetails],
        dry_run_id: uuid.UUID,
    ) -> None:
        """
        Perform a dryrun for the given resources

        :param resources: Sequence of resources for which to perform a dryrun.
        :param dry_run_id: id for this dryrun
        """
        model_version: int = resources[0].model_version
        env_id: uuid.UUID = resources[0].env_id

        async with self.cache(model_version):
            for resource in resources:
                try:
                    resource_obj: Resource | None = await self.deserialize(resource, const.ResourceAction.dryrun)
                except Exception:
                    await self.client.dryrun_update(tid=env_id, id=dry_run_id, resource=resource.rvid, changes={})
                    continue
                assert resource_obj is not None
                ctx = handler.HandlerContext(resource_obj, True)
                started = datetime.datetime.now().astimezone()
                provider = None

                resource_id: ResourceVersionIdStr = resource.rvid

                try:
                    self.logger.debug("Running dryrun for %s", resource_id)

                    try:
                        provider = await self.get_provider(resource_obj)
                    except Exception as e:
                        ctx.exception(
                            "Unable to find a handler for %(resource_id)s (exception: %(exception)s",
                            resource_id=resource_id,
                            exception=str(e),
                        )
                        await self.client.dryrun_update(
                            tid=env_id,
                            id=dry_run_id,
                            resource=resource_id,
                            changes={"handler": {"current": "FAILED", "desired": "Unable to find a handler"}},
                        )
                    else:
                        try:
                            await asyncio.get_running_loop().run_in_executor(
                                self.thread_pool, provider.execute, ctx, resource_obj, True
                            )

                            changes = ctx.changes
                            if changes is None:
                                changes = {}
                            if ctx.status == const.ResourceState.failed:
                                changes["handler"] = AttributeStateChange(current="FAILED", desired="Handler failed")
                            await self.client.dryrun_update(tid=env_id, id=dry_run_id, resource=resource_id, changes=changes)
                        except Exception as e:
                            ctx.exception(
                                "Exception during dryrun for %(resource_id)s (exception: %(exception)s",
                                resource_id=resource.rvid,
                                exception=str(e),
                            )
                            changes = ctx.changes
                            if changes is None:
                                changes = {}
                            changes["handler"] = AttributeStateChange(current="FAILED", desired="Handler failed")
                            await self.client.dryrun_update(tid=env_id, id=dry_run_id, resource=resource_id, changes=changes)

                except Exception:
                    ctx.exception("Unable to process resource for dryrun.")
                    changes = {}
                    changes["handler"] = AttributeStateChange(current="FAILED", desired="Resource Deserialization Failed")
                    await self.client.dryrun_update(tid=env_id, id=dry_run_id, resource=resource_id, changes=changes)
                finally:
                    if provider is not None:
                        provider.close()

                    finished = datetime.datetime.now().astimezone()
                    await self.client.resource_action_update(
                        tid=env_id,
                        resource_ids=[resource_id],
                        action_id=ctx.action_id,
                        action=const.ResourceAction.dryrun,
                        started=started,
                        finished=finished,
                        messages=ctx.logs,
                        status=const.ResourceState.dry,
                    )

    async def get_facts(self, resource: ResourceDetails) -> Apireturn:
        """
        Get facts for a given resource
        :param resource: The resource for which to get facts.
        """
        model_version: int = resource.model_version
        env_id: uuid.UUID = resource.env_id

        provider = None
        try:
            try:
                resource_obj: Resource | None = await self.deserialize(resource, const.ResourceAction.getfact)
            except Exception:
                return 500
            assert resource_obj is not None
            ctx = handler.HandlerContext(resource_obj)

            async with self.cache(model_version):
                try:
                    started = datetime.datetime.now().astimezone()
                    provider = await self.get_provider(resource_obj)
                    result = await asyncio.get_running_loop().run_in_executor(
                        self.thread_pool, provider.check_facts, ctx, resource_obj
                    )

                    parameters = [
                        {
                            "id": name,
                            "value": value,
                            "resource_id": resource.rid,
                            "source": ParameterSource.fact.value,
                        }
                        for name, value in result.items()
                    ]
                    # Add facts set via the set_fact() method of the HandlerContext
                    parameters.extend(ctx.facts)

                    await self.client.set_parameters(tid=env_id, parameters=parameters)
                    finished = datetime.datetime.now().astimezone()
                    await self.client.resource_action_update(
                        tid=env_id,
                        resource_ids=[resource.rvid],
                        action_id=ctx.action_id,
                        action=const.ResourceAction.getfact,
                        started=started,
                        finished=finished,
                        messages=ctx.logs,
                    )

                except Exception:
                    self.logger.exception("Unable to retrieve fact")

        except Exception:
            self.logger.exception("Unable to find a handler for %s", resource.id)
            return 500
        finally:
            if provider is not None:
                provider.close()
        return 200

    async def open_version(self, version: int) -> None:
        """
        Open a version on the cache
        """
        self._cache.open_version(version)

    async def close_version(self, version: int) -> None:
        """
        Close a version on the cache
        """
        # Needs to run on threadpool due to finalizers?
        # https://github.com/inmanta/inmanta-core/issues/833
        self._cache.close_version(version)


class InProcessExecutorManager(executor.ExecutorManager[InProcessExecutor]):
    """
    This is the executor that provides the backward compatible behavior, conforming to the agent in ISO7.

    It spawns an InProcessExecutor and makes sure all code is installed and loadable locally.
    """

    def __init__(
        self,
        environment: uuid.UUID,
        client: inmanta.protocol.SessionClient,
        eventloop: asyncio.AbstractEventLoop,
        parent_logger: logging.Logger,
        process: "agent.Agent",
    ) -> None:
        self.environment = environment
        self.client = client
        self.eventloop = eventloop
        self.logger = parent_logger
        self.process = process

        self.executors: dict[str, InProcessExecutor] = {}
        self._creation_locks: inmanta.util.NamedLock = inmanta.util.NamedLock()

    async def stop(self) -> None:
        for child in self.executors.values():
            child.stop()

    async def start(self) -> None:
        pass

    async def stop_for_agent(self, agent_name: str) -> list[InProcessExecutor]:
        if agent_name in self.executors:
            out = self.executors[agent_name]
            del self.executors[agent_name]
            out.stop()
            return [out]
        return []

    async def join(self, thread_pool_finalizer: list[ThreadPoolExecutor], timeout: float) -> None:
        for child in self.executors.values():
            child.join(thread_pool_finalizer)

    async def get_executor(
        self, agent_name: str, agent_uri: str, code: typing.Collection[executor.ResourceInstallSpec]
    ) -> InProcessExecutor:
        """
        Retrieves an Executor for a given agent with the relevant handler code loaded in its venv.
        If an Executor does not exist for the given configuration, a new one is created.

        :param agent_name: The name of the agent for which an Executor is being retrieved or created.
        :param agent_uri: The name of the host on which the agent is running.
        :param code: Collection of ResourceInstallSpec defining the configuration for the Executor i.e.
            which resource types it can act on and all necessary information to install the relevant
            handler code in its venv.
        :return: An Executor instance
        """
        if agent_name in self.executors:
            out = self.executors[agent_name]
        else:
            async with self._creation_locks.get(agent_name):
                if agent_name in self.executors:
                    out = self.executors[agent_name]
                else:
                    out = InProcessExecutor(agent_name, agent_uri, self.environment, self.client, self.eventloop, self.logger)
                    self.executors[agent_name] = out
        assert out.uri == agent_uri
        failed_resource_types: FailedResourcesSet = await self.process.ensure_code(code)
        out.failed_resource_types = failed_resource_types

        return out



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
import os
import re
import sys
import typing
import uuid
import warnings
from collections import abc, defaultdict
from configparser import ConfigParser, Interpolation
from typing import Callable, Generic, Optional, TypeVar, Union, overload

from crontab import CronTab

LOGGER = logging.getLogger(__name__)


T = TypeVar("T")


def _normalize_name(name: str) -> str:
    return name.replace("_", "-")


def _get_from_env(section: str, name: str) -> Optional[str]:
    return os.environ.get(f"INMANTA_{section}_{name}".replace("-", "_").upper(), default=None)


class LenientConfigParser(ConfigParser):
    def optionxform(self, name: str) -> str:
        name = _normalize_name(name)
        return super().optionxform(name)


class Config:
    __instance: Optional[ConfigParser] = None
    _config_dir: Optional[str] = None  # The directory this config was loaded from
    __config_definition: dict[str, dict[str, "Option"]] = defaultdict(dict)

    @classmethod
    def get_config_options(cls) -> dict[str, dict[str, "Option"]]:
        return cls.__config_definition

    @classmethod
    def load_config(
        cls,
        min_c_config_file: Optional[str] = None,
        config_dir: Optional[str] = None,
        main_cfg_file: str = "/etc/inmanta/inmanta.cfg",
    ) -> None:
        """
        Load the configuration file
        """

        cfg_files_in_config_dir: list[str]
        if config_dir and os.path.isdir(config_dir):
            cfg_files_in_config_dir = sorted(
                [os.path.join(config_dir, f) for f in os.listdir(config_dir) if f.endswith(".cfg")]
            )
        else:
            cfg_files_in_config_dir = []

        local_dot_inmanta_cfg_files: list[str] = [os.path.expanduser("~/.inmanta.cfg"), ".inmanta", ".inmanta.cfg"]

        # Files with a higher index in the list, override config options defined by files with a lower index
        files: list[str]
        if min_c_config_file is not None:
            files = [main_cfg_file] + cfg_files_in_config_dir + local_dot_inmanta_cfg_files + [min_c_config_file]
        else:
            files = [main_cfg_file] + cfg_files_in_config_dir + local_dot_inmanta_cfg_files

        config = LenientConfigParser(interpolation=Interpolation())
        config.read(files)
        cls.__instance = config
        cls._config_dir = config_dir

    @classmethod
    def load_config_from_dict(
        cls,
        input_config: typing.Mapping[str, typing.Mapping[str, typing.Any]],
    ) -> None:
        """
        Load the configuration from a dict, used to copy config.
        Replaces all existing config.
        """
        config = LenientConfigParser(interpolation=Interpolation())
        config.read_dict(input_config)
        cls.__instance = config
        cls._config_dir = None

    @classmethod
    def config_as_dict(cls) -> typing.Mapping[str, typing.Mapping[str, typing.Any]]:
        """
        Return the config as a dict, to be used with load_config_from_dict
        """
        assert cls.__instance is not None
        return dict(cls.__instance.items())

    @classmethod
    def _get_instance(cls) -> ConfigParser:
        if cls.__instance is None:
            cls.load_config()
            assert cls.__instance is not None

        return cls.__instance

    @classmethod
    def get_instance(cls) -> ConfigParser:
        """Get the singleton instance of the ConfigParser. In case it did not load the config yet, it will be loaded."""
        return cls._get_instance()

    @classmethod
    def _reset(cls) -> None:
        cls.__instance = None
        cls._config_dir = None

    @overload
    @classmethod
    def get(cls) -> ConfigParser: ...

    @overload
    @classmethod
    def get(cls, section: str, name: str, default_value: Optional[T] = None) -> Optional[str | T]: ...

    # noinspection PyNoneFunctionAssignment
    @classmethod
    def get(cls, section: Optional[str] = None, name: Optional[str] = None, default_value: Optional[object] = None) -> object:
        """
        Get the entire config or get a value directly
        """
        if section is None:
            return cls.get_instance()

        assert name is not None
        name = _normalize_name(name)

        option: Optional[Option[object]] = cls.validate_option_request(section, name, default_value)
        return cls.get_for_option(option) if option is not None else cls._get_value(section, name, default_value)

    @classmethod
    def get_for_option(cls, option: "Option[T]") -> T:
        raw_value: str | T = cls._get_value(option.section, option.name, option.get_default_value())
        return option.validate(raw_value)

    @classmethod
    def _get_value(cls, section: str, name: str, default_value: T) -> str | T:
        cfg: ConfigParser = cls.get_instance()
        val: Optional[str] = _get_from_env(section, name)
        if val is not None:
            LOGGER.debug(f"Setting {section}:{name} was set using an environment variable")
            return val
        # Typing of this method in the sdk is not entirely accurate
        # It just returns the fallback, whatever its type
        return cfg.get(section, name, fallback=default_value)

    @classmethod
    def is_set(cls, section: str, name: str) -> bool:
        """Check if a certain config option was specified in the config file."""
        return section in cls.get_instance() and name in cls.get_instance()[section]

    @classmethod
    def getboolean(cls, section: str, name: str, default_value: Optional[bool] = None) -> bool:
        """
        Return a boolean from the configuration
        """
        cls.validate_option_request(section, name, default_value)
        return cls.get_instance().getboolean(section, name, fallback=default_value)

    @classmethod
    def set(cls, section: str, name: str, value: str) -> None:
        """
        Override a value
        """
        name = _normalize_name(name)

        if section not in cls.get_instance():
            cls.get_instance().add_section(section)
        cls.get_instance().set(section, name, value)

    @classmethod
    def register_option(cls, option: "Option") -> None:
        cls.__config_definition[option.section][option.name] = option

    @classmethod
    def validate_option_request(cls, section: str, name: str, default_value: Optional[T]) -> Optional["Option[T]"]:
        if section not in cls.__config_definition:
            LOGGER.warning("Config section %s not defined" % (section))
            # raise Exception("Config section %s not defined" % (section))
            return None
        if name not in cls.__config_definition[section]:
            LOGGER.warning("Config name %s not defined in section %s", name, section)
            # raise Exception("Config name %s not defined in section %s" % (name, section))
            return None
        opt = cls.__config_definition[section][name]
        if default_value is not None and opt.get_default_value() != default_value:
            LOGGER.warning(
                "Inconsistent default value for option %s.%s: defined as %s, got %s"
                % (section, name, opt.default, default_value)
            )

        return opt


def is_int(value: str | int) -> int:
    """int"""
    return int(value)


def is_float(value: str | float) -> float:
    """float"""
    return float(value)


def is_time(value: str | int) -> int:
    """Time, the number of seconds represented as an integer value"""
    return int(value)


def is_time_or_cron(value: str | int) -> Union[int, str]:
    """Time, the number of seconds represented as an integer value or a cron-like expression"""
    try:
        return is_time(value)
    except ValueError:
        try:
            CronTab(value)
        except ValueError as e:
            raise ValueError("Not an int or cron expression: %s" % value)
        return value


def is_bool(value: Union[bool, str]) -> bool:
    """Boolean value, represented as any of true, false, on, off, yes, no, 1, 0. (Case-insensitive)"""
    if isinstance(value, bool):
        return value
    boolean_states: abc.Mapping[str, bool] = Config.get_instance().BOOLEAN_STATES
    if value.lower() not in boolean_states:
        raise ValueError("Not a boolean: %s" % value)
    return boolean_states[value.lower()]


def is_list(value: str | list[str]) -> list[str]:
    """List of comma-separated values"""
    if isinstance(value, list):
        return value
    return [] if value == "" else [x.strip() for x in value.split(",")]


def is_lower_bounded_int(lower_bound: int) -> Callable[[str | int], int]:
    """lower-bounded int factory"""

    def inner(value: str | int) -> int:
        to_int = int(value)
        if to_int < lower_bound:
            raise ValueError(f"Value can not be lower than {lower_bound}")
        return to_int

    inner.__doc__ = f"int >= {lower_bound}"
    return inner


def is_map(map_in: str | typing.Mapping[str, str]) -> typing.Mapping[str, str]:
    """List of comma-separated key=value pairs"""
    if isinstance(map_in, typing.Mapping):
        return map_in
    map_out = {}
    if map_in is not None:
        mappings = map_in.split(",")

        for mapping in mappings:
            parts = re.split("=", mapping.strip(), 1)
            if len(parts) == 2:
                key = parts[0].strip()
                value = parts[1].strip()
                if key != "" and value != "":
                    map_out[key] = value

    return map_out


def is_str(value: str) -> str:
    """str"""
    return str(value)


def is_str_opt(value: Optional[str]) -> Optional[str]:
    """optional str"""
    if value is None:
        return None
    return str(value)


def is_uuid_opt(value: Optional[str | uuid.UUID]) -> Optional[uuid.UUID]:
    """optional uuid"""
    if value is None:
        return None
    if isinstance(value, uuid.UUID):
        return value
    return uuid.UUID(value)


def is_int_opt(value: Optional[str]) -> Optional[int]:
    """optional int"""
    if value is None:
        return None
    return int(value)


class Option(Generic[T]):
    """
    Defines an option and exposes it for use

    All config option should be defined prior to use
    For the document generator to work properly, they should be defined at the module level.

    :param section: section in the config file
    :param name: name of the option
    :param default: default value for this option
        the default value is either a value or a function.
        If it is a value, `str(default)` will be used a default value.
        If it is a function, its doc string will be used to represent the value in documentation.
        and its return value as the actual default value
    :param documentation: the documentation for this option
    :param validator: a function responsible for turning the string representation of the option into the correct type.
        Its docstring is used as representation for the type of the option.
    :param predecessor_option: The Option that was deprecated in favour of this option.
    """

    def __init__(
        self,
        section: str,
        name: str,
        default: Union[T, Callable[[], T]],
        documentation: str,
        validator: Callable[[str | T], T] = is_str,
        predecessor_option: Optional["Option"] = None,
    ) -> None:
        self.section = section
        self.name = _normalize_name(name)
        self.validator = validator
        self.documentation = documentation
        self.default = default
        self.predecessor_option = predecessor_option
        Config.register_option(self)

    def get(self) -> T:
        raw_config: ConfigParser = Config.get()
        if self.predecessor_option:
            has_deprecated_option = raw_config.has_option(self.predecessor_option.section, self.predecessor_option.name)
            has_new_option = raw_config.has_option(self.section, self.name)
            if has_deprecated_option and not has_new_option:
                warnings.warn(
                    f"Config option {self.predecessor_option.name} is deprecated. Use {self.name} instead.",
                    category=DeprecationWarning,
                )
                return self.predecessor_option.get()
        return Config.get_for_option(self)

    def get_type(self) -> Optional[str]:
        if callable(self.validator):
            return self.validator.__doc__
        return None

    def get_default_desc(self) -> str:
        defa = self.default
        if callable(defa):
            return "%s" % defa.__doc__
        else:
            return f"``{defa}``"

    def validate(self, value: str | T) -> T:
        return self.validator(value)

    def get_default_value(self) -> T:
        defa = self.default
        if callable(defa):
            return defa()
        else:
            return defa

    def set(self, value: str) -> None:
        """Only for tests"""
        Config.set(self.section, self.name, value)


def option_as_default(opt: Option[T]) -> Callable[[], T]:
    """
    Wrap an option to be used as default value
    """

    def default_func() -> T:
        return opt.get()

    default_func.__doc__ = f""":inmanta.config:option:`{opt.section}.{opt.name}`"""
    return default_func


#############################
# Config
#
# Global config options are defined here
#############################
# flake8: noqa: H904
state_dir = Option("config", "state_dir", "/var/lib/inmanta", "The directory where the server stores its state", is_str)

log_dir = Option(
    "config",
    "log_dir",
    "/var/log/inmanta",
    "The directory where the resource action log is stored and the logs of auto-started agents.",
    is_str,
)

logging_config = Option(
    section="config",
    name="logging_config",
    default=None,
    documentation="The path to the configuration file for the logging framework. This is a YAML file that follows "
    "the dictionary-schema accepted by logging.config.dictConfig(). All other log-related configuration "
    "options will be ignored when this option is set.",
    validator=is_str_opt,
)


def get_executable() -> Optional[str]:
    """``os.path.abspath(sys.argv[0])``"""
    try:
        return os.path.abspath(sys.argv[0])
    except:
        return None


def get_default_nodename() -> str:
    """``socket.gethostname()``"""
    import socket

    return socket.gethostname()


nodename = Option("config", "node-name", get_default_nodename, "Force the hostname of this machine to a specific value", is_str)
feature_file_config = Option("config", "feature-file", None, "The loacation of the inmanta feature file.", is_str_opt)


###############################
# Transport Config
###############################
class TransportConfig:
    """
    A class to register the config options for Client classes
    """

    def __init__(self, name: str, port: int = 8888) -> None:
        self.prefix = "%s_rest_transport" % name
        self.host = Option(self.prefix, "host", "localhost", "IP address or hostname of the server", is_str)
        self.port = Option(self.prefix, "port", port, "Server port", is_int)
        self.ssl = Option(self.prefix, "ssl", False, "Connect using SSL?", is_bool)
        self.ssl_ca_cert_file = Option(
            self.prefix, "ssl_ca_cert_file", None, "CA cert file used to validate the server certificate against", is_str_opt
        )
        self.token = Option(self.prefix, "token", None, "The bearer token to use to connect to the API", is_str_opt)
        self.request_timeout = Option(
            self.prefix, "request_timeout", 120, "The time before a request times out in seconds", is_int
        )
        self.max_clients = Option(
            self.prefix,
            "max_clients",
            None,
            "The maximum number of simultaneous connections that can be open in parallel",
            is_int_opt,
        )


compiler_transport = TransportConfig("compiler")
TransportConfig("client")
cmdline_rest_transport = TransportConfig("cmdline")



"""
    Copyright 2023 Inmanta
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
        http://www.apache.org/licenses/LICENSE-2.0
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    schema = """
    -- change the type of the 'discovered_at' column
    ALTER TABLE public.discoveredresource
    ALTER COLUMN discovered_at TYPE TIMESTAMP WITH TIME ZONE,
    ALTER COLUMN discovered_at SET NOT NULL;
    """
    await connection.execute(schema)



"""
    Copyright 2021 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
import os
import py_compile
import re
import shutil
import sys
from collections import abc
from collections.abc import Sequence
from typing import Optional

import py
import pytest
from pkg_resources import Requirement

from inmanta import compiler, const, env, loader, plugins, resources
from inmanta.ast import CompilerException
from inmanta.const import CF_CACHE_DIR
from inmanta.data.model import PipConfig
from inmanta.env import ConflictingRequirements, LocalPackagePath, PackageNotFound, process_env
from inmanta.module import (
    DummyProject,
    InmantaModuleRequirement,
    ModuleLoadingException,
    ModuleNotFoundException,
    ModuleV1,
    ModuleV2,
    ModuleV2Source,
    Project,
)
from inmanta.moduletool import ModuleConverter, ModuleTool, ProjectTool
from packaging.version import Version
from utils import PipIndex, create_python_package, log_contains, module_from_template, v1_module_from_template


@pytest.mark.parametrize_any("editable_install", [True, False])
def test_v2_module_loading(
    editable_install: bool,
    tmpdir: py.path.local,
    snippetcompiler,
    capsys,
    modules_v2_dir: str,
    local_module_package_index,  # upstream for setuptools for isolated build
) -> None:
    # Disable modules_dir
    snippetcompiler.modules_dir = None

    module_name = "elaboratev2module"
    module_dir = os.path.join(modules_v2_dir, module_name)
    module_copy_dir = os.path.join(tmpdir, "module")
    shutil.copytree(module_dir, module_copy_dir)
    assert os.path.isdir(module_copy_dir)

    snippetcompiler.setup_for_snippet(
        f"""
            import {module_name}

            {module_name}::print_message("Hello world")
        """,
        autostd=False,
        install_v2_modules=[LocalPackagePath(path=module_copy_dir, editable=editable_install)],
        index_url=local_module_package_index,
    )

    snippetcompiler.do_export()
    assert "Hello world" in capsys.readouterr().out

    # Make sure the cache files are created
    cache_folder = os.path.join(snippetcompiler.project_dir, CF_CACHE_DIR)
    assert len(os.listdir(cache_folder)) > 0


def test_v1_and_v2_module_installed_simultaneously(
    tmpdir: py.path.local, snippetcompiler_clean, capsys, caplog, modules_dir: str
) -> None:
    """
    When a module is installed both in V1 and V2 format, ensure that:
       * A warning is logged
       * The V2 module is loaded and not the V1 module.
    """

    module_name = "v1_print_plugin"

    def compile_and_verify(
        expected_message: str, expect_warning: bool, install_v2_modules: list[LocalPackagePath] = []
    ) -> None:
        caplog.clear()
        snippetcompiler_clean.setup_for_snippet(f"import {module_name}", install_v2_modules=install_v2_modules, autostd=False)
        snippetcompiler_clean.do_export()
        assert expected_message in capsys.readouterr().out
        got_warning = f"Module {module_name} is installed as a V1 module and a V2 module" in caplog.text
        assert got_warning == expect_warning

    # Run compile. Only a V1 module is installed in the module path
    expected_message_v1 = "Hello world"
    compile_and_verify(expected_message=expected_message_v1, expect_warning=False)
    assert isinstance(Project.get().modules[module_name], ModuleV1)

    # Convert V1 module to V2 module and install it as well
    module_dir = os.path.join(modules_dir, module_name)
    v1_module_dir = os.path.join(tmpdir, "v1_module")
    shutil.copytree(module_dir, v1_module_dir)
    assert os.path.isdir(v1_module_dir)
    v2_module_dir = os.path.join(tmpdir, "v2_module")
    module = ModuleV1(project=DummyProject(autostd=False), path=v1_module_dir)
    ModuleConverter(module).convert(output_directory=v2_module_dir)

    # Print a different message in the V2 module, to detect which of both gets loaded
    expected_message_v2 = "Other message"
    with open(os.path.join(v2_module_dir, "model", "_init.cf"), "r+") as fd:
        content = fd.read()
        assert expected_message_v1 in content
        content = content.replace(expected_message_v1, expected_message_v2)
        assert expected_message_v2 in content
        fd.seek(0)
        fd.write(content)

    # Run compile again. V1 version and V2 version are installed simultaneously
    compile_and_verify(
        expected_message=expected_message_v2,
        expect_warning=True,
        install_v2_modules=[LocalPackagePath(path=v2_module_dir, editable=False)],
    )
    assert isinstance(Project.get().modules[module_name], ModuleV2)


def test_v1_module_depends_on_v1_and_v2_module(
    tmpdir: py.path.local, snippetcompiler, capsys, modules_v2_dir: str, caplog
) -> None:
    module_dir = os.path.join(modules_v2_dir, "elaboratev2module")
    module_copy_dir = os.path.join(tmpdir, "module")
    shutil.copytree(module_dir, module_copy_dir)
    assert os.path.isdir(module_copy_dir)

    module_name = "v1_module_depends_on_v1_and_v2_module"
    snippetcompiler.setup_for_snippet(
        f"import {module_name}",
        autostd=False,
        install_v2_modules=[LocalPackagePath(path=module_copy_dir, editable=False)],
    )

    snippetcompiler.do_export()
    output = capsys.readouterr().out
    assert "Print from v2 module" in output
    assert "Hello world" in output

    # Assert deprecation warnings regarding the use of V1 modules
    for v1_module_name in [module_name, "elaboratev1module"]:
        log_contains(
            caplog,
            "inmanta.warnings",
            logging.WARNING,
            f"Loaded V1 module {v1_module_name}. The use of V1 modules is deprecated. Use the equivalent V2 module instead.",
        )


def test_install_module_no_v2_source(snippetcompiler) -> None:
    """
    Verify that attempting to install a v2 module without a v2 module source configured results in the appropriate exception.
    """
    module_name = "non_existing_module"
    with pytest.raises(PackageNotFound) as e:
        snippetcompiler.setup_for_snippet(
            snippet=f"import {module_name}",
            install_project=True,
            python_requires=[
                InmantaModuleRequirement.parse(module_name).get_python_package_requirement(),
            ],
        )

    message: str = (
        "Attempting to install a v2 module non_existing_module but pip is not configured. Add the relevant pip indexes "
        "to the project config file. e.g. to set PyPi as pip index, add the following to `project.yml`:\n"
        "pip:\n"
        "  index_url: https://pypi.org/simple\n"
        "Another option is to set `pip.use_system_config = true` to use the system's pip config."
    )
    e.match(re.escape(message))


@pytest.mark.parametrize("allow_v1", [True, False])
def test_load_module_v1_already_installed(snippetcompiler, modules_dir: str, allow_v1: bool) -> None:
    """
    Test whether the Project.load_module() method works correctly when loading a V1 module that was already installed
    in the module path.
    """
    module_name = "elaboratev1module"
    module_dir = os.path.join(modules_dir, module_name)
    project: Project = snippetcompiler.setup_for_snippet(
        snippet=f"import {module_name}", add_to_module_path=[module_dir], install_project=False
    )

    assert module_name not in project.modules
    if allow_v1:
        project.load_module(module_name=module_name, install_v1=False, allow_v1=allow_v1)
        assert module_name in project.modules
    else:
        with pytest.raises(ModuleNotFoundException, match=f"Could not find module {module_name}"):
            project.load_module(module_name=module_name, install_v1=False, allow_v1=allow_v1)


def test_load_module_v1_module_using_install(snippetcompiler) -> None:
    """
    Test whether the Project.load_module() method works correctly when a module is only available as a V1 module
    and that module is not yet present in the module path.
    """
    module_name = "std"
    project: Project = snippetcompiler.setup_for_snippet(snippet=f"import {module_name}", install_project=False)
    # Remove std module in downloadpath created by other test case
    shutil.rmtree(os.path.join(project.downloadpath, module_name), ignore_errors=True)
    assert module_name not in project.modules
    assert module_name not in os.listdir(project.downloadpath)
    project.load_module(module_name=module_name, install_v1=True, allow_v1=True)
    assert module_name in project.modules
    assert module_name in os.listdir(project.downloadpath)


@pytest.mark.parametrize("allow_v1", [True, False])
@pytest.mark.parametrize("editable_install", [True, False])
def test_load_module_v2_already_installed(
    snippetcompiler_clean,
    modules_v2_dir: str,
    allow_v1: bool,
    editable_install: bool,
    local_module_package_index,
) -> None:
    """
    Test whether the Project.load_module() method works correctly when loading a V2 module that was already installed
    in the compiler venv.
    """
    module_name = "elaboratev2module"
    module_dir = os.path.join(modules_v2_dir, module_name)
    project: Project = snippetcompiler_clean.setup_for_snippet(
        snippet=f"import {module_name}",
        install_v2_modules=[LocalPackagePath(module_dir, editable_install)],
        install_project=False,
        index_url=local_module_package_index,
    )

    assert module_name not in project.modules
    project.load_module(module_name=module_name, install_v2=False, allow_v1=allow_v1)
    assert module_name in project.modules


@pytest.mark.parametrize("install", [True, False])
def test_load_module_v2_module_using_install(
    snippetcompiler_clean,
    local_module_package_index: str,
    install: bool,
):
    """
    Test whether the Project.load_module() method works correctly when a module is only available as a V2 module
    and that module is not yet installed in the compiler venv.
    """
    module_name = "minimalv2module"
    project: Project = snippetcompiler_clean.setup_for_snippet(
        snippet=f"import {module_name}", index_url=local_module_package_index, install_project=False
    )
    assert module_name not in project.modules
    assert module_name not in os.listdir(project.downloadpath)
    if install:
        project.load_module(module_name=module_name, install_v1=install, install_v2=install, allow_v1=True)
    else:
        with pytest.raises(ModuleNotFoundException, match=f"Could not find module {module_name}"):
            project.load_module(module_name=module_name, install_v1=install, install_v2=install, allow_v1=True)
    assert (module_name in project.modules) == install
    assert module_name not in os.listdir(project.downloadpath)


@pytest.mark.parametrize("allow_v1", [True, False])
def test_load_module_module_not_found(snippetcompiler_clean, allow_v1: bool) -> None:
    """
    Assert behavior when a module is not found as a V1 or a V2 module.
    """
    module_name = "non_existing_module"
    snippetcompiler_clean.modules_dir = None
    project: Project = snippetcompiler_clean.setup_for_snippet(
        snippet=f"import {module_name}", install_project=False, index_url="non_existing_local_index"
    )
    with pytest.raises(ModuleNotFoundException, match=f"Could not find module {module_name}"):
        project.load_module(module_name=module_name, install_v1=allow_v1, install_v2=True, allow_v1=allow_v1)


def test_load_module_v1_and_v2_installed(
    tmpdir: py.path.local,
    snippetcompiler_clean,
    modules_dir: str,
    caplog,
) -> None:
    """
    Test whether the Project.load_module() method works correctly when the v1 and v2 version of a module are both installed.
    The V2 module should be loaded and a warning should be logged.
    """
    module_name = "minimalv1module"
    module_dir = os.path.join(modules_dir, module_name)

    # Convert v1 to v2 module
    module = ModuleV1(project=DummyProject(autostd=False), path=module_dir)
    module_dir_v2 = os.path.join(tmpdir, f"{module_name}-v2")
    ModuleConverter(module).convert(output_directory=module_dir_v2)

    # The V1 module is installed implicitly. The snippetcompiler_clean fixture adds
    # the `modules_dir` to the modulepath of the newly created project.
    project: Project = snippetcompiler_clean.setup_for_snippet(
        snippet=f"import {module_name}",
        install_v2_modules=[LocalPackagePath(module_dir_v2, editable=False)],
        install_project=False,
    )

    assert module_name not in project.modules
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        project.load_module(module_name=module_name, install_v1=False, install_v2=False, allow_v1=True)
    assert f"Module {module_name} is installed as a V1 module and a V2 module: V1 will be ignored." in caplog.text
    assert module_name in project.modules
    assert isinstance(project.modules[module_name], ModuleV2)


@pytest.mark.parametrize("preload_v1_module", [True, False])
def test_load_module_recursive_v2_module_depends_on_v1(
    local_module_package_index: str, snippetcompiler, preload_v1_module: bool
) -> None:
    """
    A V2 module cannot depend on a V1 module. This test case ensure that the load_module_recursive() method
    raises an error when a dependency of a V2 module is only available as a V1 module.

    Dependency graph:  v2_depends_on_v1 (V2)  --->  mod1 (V1)
    """
    project = snippetcompiler.setup_for_snippet(
        snippet="import v2_depends_on_v1",
        index_url=local_module_package_index,
        python_requires=[Requirement.parse("inmanta-module-v2-depends-on-v1")],
        install_project=False,
    )
    if preload_v1_module:
        project.get_module("mod1", allow_v1=True)
    assert ("mod1" in project.modules) == preload_v1_module

    with pytest.raises(ModuleLoadingException, match="Failed to load module mod1"):
        project.load_module_recursive(install=True)


def test_load_module_recursive_complex_module_dependencies(local_module_package_index: str, snippetcompiler) -> None:
    """
    Test whether the load_module_recursive() method works correctly when complex, circular dependencies exist between modules.

    Dependency graph:

    complex_module_dependencies_mod1  --------------------->  complex_module_dependencies_mod2
           |   ^                      <---------------------             |   ^
           |   |                                                         |   |
           v   |                                                         v   |
    complex_module_dependencies_mod1::submod                  complex_module_dependencies_mod2::submod
    """
    project = snippetcompiler.setup_for_snippet(
        snippet="import complex_module_dependencies_mod1",
        autostd=False,
        index_url=local_module_package_index,
        python_requires=[Requirement.parse("inmanta-module-complex-module-dependencies-mod1")],
        install_project=False,
    )
    assert "complex_module_dependencies_mod1" not in project.modules
    assert "complex_module_dependencies_mod2" not in project.modules
    loaded_namespaces: set[str] = {ns for ns, _, _ in project.load_module_recursive(install=True)}
    assert "complex_module_dependencies_mod1" in project.modules
    assert "complex_module_dependencies_mod2" in project.modules
    expected_namespaces = {
        "complex_module_dependencies_mod1",
        "complex_module_dependencies_mod1::submod",
        "complex_module_dependencies_mod2",
        "complex_module_dependencies_mod2::submod",
    }
    assert loaded_namespaces == expected_namespaces


def test_load_import_based_v2_project(local_module_package_index: str, snippetcompiler_clean) -> None:
    """
    A project needs to explicitly list its v2 dependencies in order to be able to load them. Import-based loading is not
    allowed.
    """
    module_name: str = "minimalv2module"

    def load(requires: Optional[list[Requirement]] = None) -> None:
        project: Project = snippetcompiler_clean.setup_for_snippet(
            f"import {module_name}",
            autostd=False,
            install_project=False,
            index_url=local_module_package_index,
            # make sure that even listing the requirement in project.yml does not suffice
            project_requires=[InmantaModuleRequirement.parse(module_name)],
            python_requires=requires,
        )
        project.load_module_recursive(install=True)

    with pytest.raises(ModuleLoadingException, match=f"Failed to load module {module_name}"):
        load()
    # assert that it doesn't raise an error with explicit requirements set
    load([Requirement.parse(ModuleV2Source.get_package_name_for(module_name))])


@pytest.mark.parametrize("v1", [True, False])
@pytest.mark.parametrize("explicit_dependency", [True, False])
def test_load_import_based_v2_module(
    local_module_package_index: str,
    snippetcompiler_clean,
    modules_dir: str,
    modules_v2_dir: str,
    tmpdir: py.path.local,
    v1: bool,
    explicit_dependency: bool,
) -> None:
    """
    A module needs to explicitly list its v2 dependencies in order to be able to load them. Import-based loading is not
    allowed.
    """
    main_module_name: str = "mymodule"
    dependency_module_name: str = "minimalv2module"
    index: PipIndex = PipIndex(os.path.join(str(tmpdir), ".my-index"))
    libs_dir: str = os.path.join(str(tmpdir), "libs")
    os.makedirs(libs_dir)

    model: str = f"import {dependency_module_name}"
    requirements: list[InmantaModuleRequirement] = (
        [InmantaModuleRequirement.parse(dependency_module_name)] if explicit_dependency else []
    )

    if v1:
        v1_module_from_template(
            os.path.join(modules_dir, "minimalv1module"),
            os.path.join(libs_dir, main_module_name),
            new_name=main_module_name,
            new_content_init_cf=model,
            new_requirements=requirements,
        )
    else:
        module_from_template(
            os.path.join(modules_v2_dir, "minimalv2module"),
            os.path.join(str(tmpdir), main_module_name),
            new_name=main_module_name,
            new_content_init_cf=model,
            new_requirements=requirements,
            install=False,
            publish_index=index,
        )

    project: Project = snippetcompiler_clean.setup_for_snippet(
        f"import {main_module_name}",
        autostd=False,
        install_project=False,
        add_to_module_path=[libs_dir],
        index_url=local_module_package_index,
        extra_index_url=[index.url],
        # make sure that even listing the requirement in project.yml does not suffice
        project_requires=[InmantaModuleRequirement.parse(dependency_module_name)],
        python_requires=[] if v1 else [Requirement.parse(ModuleV2Source.get_package_name_for(main_module_name))],
    )

    if explicit_dependency:
        # assert that it doesn't raise an error with explicit requirements set
        project.load_module_recursive(install=True)
    else:
        with pytest.raises(ModuleLoadingException, match=f"Failed to load module {dependency_module_name}"):
            project.load_module_recursive(install=True)


def test_module_unload(
    local_module_package_index: str,
    snippetcompiler,
) -> None:
    """
    Verify that Module.unload removes it from the project, unloads its Python modules and deregisters its plugins.
    """
    project: Project = snippetcompiler.setup_for_snippet(
        """
import minimalv2module
import elaboratev2module
        """.strip(),
        index_url=local_module_package_index,
        python_requires=[
            InmantaModuleRequirement.parse("minimalv2module").get_python_package_requirement(),
            InmantaModuleRequirement.parse("elaboratev2module").get_python_package_requirement(),
        ],
        autostd=False,
    )
    project.load()

    assert "minimalv2module" in project.modules
    assert "elaboratev2module" in project.modules

    assert "inmanta_plugins.minimalv2module" in sys.modules
    assert "inmanta_plugins.elaboratev2module" in sys.modules

    assert "elaboratev2module::print_message" in plugins.PluginMeta.get_functions()

    project.modules["elaboratev2module"].unload()

    assert "minimalv2module" in project.modules
    assert "elaboratev2module" not in project.modules

    assert "inmanta_plugins.minimalv2module" in sys.modules
    assert "inmanta_plugins.elaboratev2module" not in sys.modules

    assert "elaboratev2module::print_message" not in plugins.PluginMeta.get_functions()


def test_project_has_v2_requirements_on_non_imported_module(
    snippetcompiler,
    local_module_package_index: str,
) -> None:
    """
    A Project has a module V2 requirement on a module that it doesn't import.
    Ensure that the non-imported module is not loaded.
    """
    dependency = "elaboratev2module"
    project: Project = snippetcompiler.setup_for_snippet(
        snippet="",  # Don't import elaboratev2module
        index_url=local_module_package_index,
        python_requires=[
            InmantaModuleRequirement.parse(dependency).get_python_package_requirement(),
        ],
        autostd=False,
    )
    project.load_module_recursive()
    assert dependency not in project.modules


def test_module_has_v2_requirements_on_non_imported_module(snippetcompiler, local_module_package_index: str) -> None:
    """
    A module has a module V2 requirement on a module that it doesn't import.
    Ensure that the non-imported module is not loaded.

    Scenario: module dependency_but_no_import has dependency on minimalv2module
              but dependency_but_no_import doesn't import minimalv2module.
    """
    project: Project = snippetcompiler.setup_for_snippet(
        snippet="import dependency_but_no_import",
        index_url=local_module_package_index,
        python_requires=[
            InmantaModuleRequirement.parse("dependency_but_no_import").get_python_package_requirement(),
        ],
        autostd=False,
    )
    project.load_module_recursive()
    assert "minimalv2module" not in project.modules


@pytest.mark.slowtest
def test_module_v2_load_installed_without_required(snippetcompiler_clean, local_module_package_index: str) -> None:
    """
    Verify that module loading works as expected for a v2 module installed in the environment while it is not in any of the
    requirements files. This is relevant for correct working of pytest-inmanta's auto-generated projects.
    """
    # set up venv
    snippetcompiler_clean.setup_for_snippet("", autostd=False)
    process_env.install_for_config(
        [InmantaModuleRequirement.parse("elaboratev2module").get_python_package_requirement()],
        config=PipConfig(
            index_url=local_module_package_index,
        ),
    )

    # set up project: import submodule only, don't add the module to requirements
    project: Project = snippetcompiler_clean.setup_for_snippet("import elaboratev2module::nested::sub", autostd=False)
    assert {tup[0] for tup in project.load_module_recursive()} == {
        "elaboratev2module",
        "elaboratev2module::nested",
        "elaboratev2module::nested::sub",
    }


@pytest.mark.slowtest
def test_project_requirements_dont_overwrite_core_requirements_source(
    snippetcompiler_clean,
    local_module_package_index: str,
    modules_v2_dir: str,
    tmpdir: py.path.local,
) -> None:
    """
    A project has a requirement that is also a requirement of core
    but with another version. The requirements of core should not be
    overwritten. The module gets installed from source
    """
    if "inmanta-core" in process_env.get_installed_packages(only_editable=True):
        pytest.skip(
            "This test would fail if it runs against an inmanta-core installed in editable mode, because the build tag "
            "on the development branch is set to .dev0. The inmanta package protection feature would make pip "
            "install a non-editable version of the same package. But no version with build tag .dev0 exists on the python "
            "package repository."
        )

    # Create the module
    module_name: str = "minimalv2module"
    module_path: str = str(tmpdir.join(module_name))
    module_from_template(
        os.path.join(modules_v2_dir, module_name), module_path, new_requirements=[Requirement.parse("Jinja2==2.11.3")]
    )

    # Activate the snippetcompiler venv
    project: Project = snippetcompiler_clean.setup_for_snippet("")
    active_env = project.virtualenv
    jinja2_version_before = active_env.get_installed_packages()["Jinja2"].base_version

    # Install the module
    mod_artifact_path = ModuleTool().build(path=module_path)
    with pytest.raises(ConflictingRequirements) as e:
        env.process_env.install_from_source([env.LocalPackagePath(path=mod_artifact_path, editable=False)])

    assert ("these package versions have conflicting dependencies") in str(e.value.msg)

    jinja2_version_after = active_env.get_installed_packages()["Jinja2"].base_version
    assert jinja2_version_before == jinja2_version_after


@pytest.mark.slowtest
def test_project_requirements_dont_overwrite_core_requirements_index(
    snippetcompiler_clean,
    modules_v2_dir: str,
    tmpdir: py.path.local,
) -> None:
    """
    A module from index has a requirement that is also a requirement of core
    but with another version. The requirements of core should not be
    overwritten. The module gets installed from index.
    """
    if "inmanta-core" in process_env.get_installed_packages(only_editable=True):
        pytest.skip(
            "This test would fail if it runs against an inmanta-core installed in editable mode, because the build tag "
            "on the development branch is set to .dev0. The inmanta package protection feature would make pip "
            "install a non-editable version of the same package. But no version with build tag .dev0 exists on the python "
            "package repository."
        )
    # Create the module
    module_name: str = "minimalv2module"
    module_path: str = str(tmpdir.join(module_name))
    index: PipIndex = PipIndex(artifact_dir=os.path.join(str(tmpdir), ".custom-index"))
    module_from_template(
        os.path.join(modules_v2_dir, module_name),
        module_path,
        new_requirements=[Requirement.parse("Jinja2==2.11.3")],
        publish_index=index,
    )

    # Setup project
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "",
        install_project=False,
        index_url=index.url,
        extra_index_url=["https://pypi.org/simple"],
        python_requires=[InmantaModuleRequirement.parse(module_name).get_python_package_requirement()],
        autostd=False,
    )

    active_env = project.virtualenv
    jinja2_version_before = active_env.get_installed_packages()["Jinja2"].base_version

    # Install project
    with pytest.raises(ConflictingRequirements):
        project.install_modules()

    jinja2_version_after = active_env.get_installed_packages()["Jinja2"].base_version
    assert jinja2_version_before == jinja2_version_after


@pytest.mark.slowtest
@pytest.mark.parametrize("strict_deps_check", [True, False, None])
def test_module_conflicting_dependencies_with_v2_modules(
    snippetcompiler_clean,
    modules_v2_dir: str,
    tmpdir: py.path.local,
    caplog,
    strict_deps_check: Optional[bool],
) -> None:
    """
    Show an error message when installing a module that breaks the dependencies
    of another one. minimalv2module depends on y~=1.0.0 which requires x~=1.0.0.
    after the install of minimalv2module we try to install minimalv2module2 which
    requires x~=2.0.0. The y~=1.0.0 requirement is now broken as python package x
    has now version 2.0.0 and y needs 1.0.0.
    """
    index: PipIndex = PipIndex(artifact_dir=os.path.join(str(tmpdir), ".custom-index"))

    # Create a python package x with version 1.0.0
    create_python_package("x", Version("1.0.0"), str(tmpdir.join("x-1.0.0")), publish_index=index)

    # Create a python package x with version 2.0.0
    create_python_package("x", Version("2.0.0"), str(tmpdir.join("x-2.0.0")), publish_index=index)

    # Create a python package y with version 1.0.0 that depends on x~=1.0.0
    create_python_package(
        "y", Version("1.0.0"), str(tmpdir.join("y-1.0.0")), requirements=[Requirement.parse("x~=1.0.0")], publish_index=index
    )

    # Create the first module
    module_name1: str = "minimalv2module"
    module_path1: str = str(tmpdir.join(module_name1))
    module_from_template(
        os.path.join(modules_v2_dir, module_name1),
        module_path1,
        new_requirements=[Requirement.parse("y~=1.0.0")],
        publish_index=index,
    )

    # Create the second module
    module_name2: str = "minimalv2module2"
    module_path2: str = str(tmpdir.join(module_name2))
    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        module_path2,
        new_name="minimalv2module2",
        new_requirements=[Requirement.parse("x~=2.0.0")],
        publish_index=index,
    )

    req1 = InmantaModuleRequirement.parse(module_name1).get_python_package_requirement()
    req2 = InmantaModuleRequirement.parse(module_name2).get_python_package_requirement()
    # Setup project
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "",
        install_project=False,
        index_url=index.url,
        python_requires=[req1, req2],
        autostd=False,
        strict_deps_check=strict_deps_check,
    )

    # Install project
    msg: str = "Incompatibility between constraint x"
    if strict_deps_check is None or strict_deps_check:
        with pytest.raises(ConflictingRequirements) as e:
            project.install_modules()
        assert msg in e.value.get_message()
    else:
        # The version conflict is present in a transitive dependency, so without strict_deps_check enabled,
        # only a warning message will be logged.
        with caplog.at_level(logging.WARNING):
            caplog.clear()
            project.install_modules()
            assert msg in caplog.text


@pytest.mark.slowtest
@pytest.mark.parametrize("strict_deps_check", [True, False, None])
def test_module_conflicting_dependencies_with_v1_module(
    snippetcompiler_clean,
    modules_dir: str,
    modules_v2_dir: str,
    tmpdir: py.path.local,
    strict_deps_check: Optional[bool],
) -> None:
    """
    Show an error message when installing a module that breaks the dependencies
    of another one. modulev1 depends on y~=1.0.0.
    after the install of modulev1 we try to install minimalv2module2 which
    requires y~=2.0.0. those 2 requirements conflict which each other.
    """
    index: PipIndex = PipIndex(artifact_dir=os.path.join(str(tmpdir), ".custom-index"))
    # Create a python package x with version 1.0.0
    create_python_package("y", Version("1.0.0"), str(tmpdir.join("y-1.0.0")), publish_index=index)

    # Create a python package x with version 2.0.0
    create_python_package("y", Version("2.0.0"), str(tmpdir.join("y-2.0.0")), publish_index=index)

    # Create the first module
    module_name1: str = "minimalv1module"
    module_path1: str = str(tmpdir.join("modulev1"))
    v1_module_from_template(
        os.path.join(modules_dir, module_name1),
        module_path1,
        new_name="modulev1",
        new_requirements=[Requirement.parse("y~=1.0.0")],
    )

    # Create the second module
    module_name2: str = "minimalv2module"
    module_path2: str = str(tmpdir.join(module_name2))
    module_from_template(
        os.path.join(modules_v2_dir, module_name2),
        module_path2,
        new_requirements=[Requirement.parse("y~=2.0.0")],
        publish_index=index,
    )

    req = InmantaModuleRequirement.parse(module_name2).get_python_package_requirement()

    # Setup project
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "import modulev1",
        install_project=False,
        index_url=index.url,
        python_requires=[req],
        autostd=False,
        add_to_module_path=[str(tmpdir)],
        strict_deps_check=strict_deps_check,
    )

    # Install project
    with pytest.raises(CompilerException) as e:
        # The version conflict is present in a direct dependency, so this always results in an error.
        project.install_modules()
    if strict_deps_check is None or strict_deps_check:
        assert "Incompatibility between constraint y" in e.value.get_message()
    else:
        assert "requirements conflicts were found" in e.value.get_message()


@pytest.mark.slowtest
@pytest.mark.parametrize("package_name_extra", ["inmanta-module-minimalv2module", "lorem"])
def test_module_install_extra_on_project_level_v2_dep(
    snippetcompiler_clean,
    modules_v2_dir: str,
    tmpdir: py.path.local,
    local_module_package_index: str,
    package_name_extra: str,
) -> None:
    """
    Verify that module installation works correctly when a project has a V2 module dependency with an extra.
    """
    index: PipIndex = PipIndex(artifact_dir=str(tmpdir.join(".index")))

    # create module with optional dependency
    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("mymod")),
        new_name="mymod",
        new_requirements=[],
        new_extras={
            "myfeature": [Requirement.parse(package_name_extra)],
        },
        publish_index=index,
    )
    package_with_extra: Requirement = InmantaModuleRequirement.parse("mymod[myfeature]").get_python_package_requirement()
    package_name: str = f"{ModuleV2.PKG_NAME_PREFIX}mymod"

    # project with dependency on mymod with extra
    snippetcompiler_clean.setup_for_snippet(
        "import mymod",
        install_project=True,
        index_url=index.url,
        extra_index_url=[local_module_package_index, "https://pypi.org/simple"],
        python_requires=[package_with_extra],
        autostd=False,
    )

    installed: abc.Mapping[str, Version] = process_env.get_installed_packages()
    assert package_name in installed
    assert package_name_extra in installed


@pytest.mark.slowtest
@pytest.mark.parametrize("package_name_extra", ["inmanta-module-minimalv2module", "lorem"])
def test_module_install_extra_on_dep_of_v2_module(
    snippetcompiler_clean,
    modules_v2_dir: str,
    tmpdir: py.path.local,
    local_module_package_index: str,
    package_name_extra: str,
) -> None:
    """
    Verify that module installation works correctly when a V2 module defines a dependency with an extra.

    Dependency tree:

        myv2mod
           --> depmod (V2)
           [--> <extra>]
    """
    index: PipIndex = PipIndex(artifact_dir=str(tmpdir.join(".index")))

    # create module with optional dependency
    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("depmod")),
        new_name="depmod",
        new_requirements=[],
        new_extras={
            "myfeature": [Requirement.parse(package_name_extra)],
        },
        publish_index=index,
    )

    # Create myv2mod with extra
    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("myv2mod")),
        new_name="myv2mod",
        new_content_init_cf="import depmod",
        new_requirements=[InmantaModuleRequirement.parse("depmod[myfeature]").get_python_package_requirement()],
        publish_index=index,
    )
    # project with dependency on myv2mod with extra
    snippetcompiler_clean.setup_for_snippet(
        "import myv2mod",
        install_project=True,
        index_url=index.url,
        extra_index_url=[local_module_package_index, "https://pypi.org/simple"],
        python_requires=[Requirement.parse("inmanta-module-myv2mod")],
        autostd=False,
    )

    installed: abc.Mapping[str, Version] = process_env.get_installed_packages()
    package_name: str = f"{ModuleV2.PKG_NAME_PREFIX}depmod"
    assert package_name in installed
    assert package_name_extra in installed


@pytest.mark.slowtest
@pytest.mark.parametrize("package_name_extra", ["inmanta-module-minimalv2module", "lorem"])
def test_module_install_extra_on_dep_of_v1_module(
    tmpdir,
    snippetcompiler_clean,
    modules_dir: str,
    modules_v2_dir: str,
    local_module_package_index: str,
    package_name_extra: str,
) -> None:
    """
    Verify that module installation works correct when a V1 module has a dependency that defines an extra.

    Dependency tree:

        myv1mod
           --> depmod (V2)
           [--> <extra>]
    """
    index: PipIndex = PipIndex(artifact_dir=str(tmpdir.join(".index")))

    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("depmod")),
        new_name="depmod",
        new_requirements=[],
        new_extras={
            "myfeature": [Requirement.parse(package_name_extra)],
        },
        publish_index=index,
    )

    # v1 module with dependency on myv1mod with extra
    v1_module_from_template(
        os.path.join(modules_dir, "minimalv1module"),
        str(tmpdir.join("myv1mod")),
        new_name="myv1mod",
        new_requirements=[InmantaModuleRequirement.parse("depmod[myfeature]").get_python_package_requirement()],
    )
    snippetcompiler_clean.setup_for_snippet(
        "import myv1mod",
        install_project=True,
        add_to_module_path=[str(tmpdir)],
        index_url=index.url,
        extra_index_url=[local_module_package_index, "https://pypi.org/simple"],
        autostd=False,
    )

    installed: abc.Mapping[str, Version] = process_env.get_installed_packages()
    package_name: str = f"{ModuleV2.PKG_NAME_PREFIX}depmod"
    assert package_name in installed
    assert package_name_extra in installed


@pytest.mark.slowtest
@pytest.mark.parametrize("package_name_extra", ["inmanta-module-minimalv2module", "lorem"])
@pytest.mark.parametrize("do_project_update", [True, False])
def test_module_install_extra_on_project_level_v2_dep_update_scenario(
    snippetcompiler_clean,
    modules_v2_dir: str,
    tmpdir: py.path.local,
    local_module_package_index: str,
    package_name_extra: str,
    do_project_update: bool,
) -> None:
    """
    Verify that module installation works correctly when a project is updated with a V2 module dependency with an extra.
    """
    index: PipIndex = PipIndex(artifact_dir=str(tmpdir.join(".index")))

    # create module with optional dependency
    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("mymod")),
        new_name="mymod",
        new_requirements=[],
        new_extras={
            "myfeature": [Requirement.parse(package_name_extra)],
        },
        publish_index=index,
    )
    package_without_extra: Requirement = InmantaModuleRequirement.parse("mymod").get_python_package_requirement()
    package_with_extra: Requirement = InmantaModuleRequirement.parse("mymod[myfeature]").get_python_package_requirement()
    package_name: str = str(package_without_extra)

    def assert_installed(*, module_installed: bool = True, extra_installed: bool) -> None:
        installed: abc.Mapping[str, Version] = process_env.get_installed_packages()
        assert (package_name in installed) == module_installed
        assert (package_name_extra in installed) == extra_installed

    # project with dependency on mymod without extra
    snippetcompiler_clean.setup_for_snippet(
        "import mymod",
        install_project=True,
        index_url=index.url,
        extra_index_url=[local_module_package_index, "https://pypi.org/simple"],
        python_requires=[package_without_extra],
        autostd=False,
    )
    assert_installed(extra_installed=False)

    # project with dependency on mymod with extra
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "import mymod",
        install_project=not do_project_update,
        index_url=index.url,
        extra_index_url=[local_module_package_index, "https://pypi.org/simple"],
        python_requires=[package_with_extra],
        autostd=False,
    )
    if do_project_update:
        project_tool = ProjectTool()
        project_tool.update(project=project)

    assert_installed(extra_installed=True)


@pytest.mark.slowtest
@pytest.mark.parametrize("package_name_extra", ["inmanta-module-minimalv2module", "lorem"])
@pytest.mark.parametrize("do_project_update", [True, False])
def test_module_install_extra_on_dep_of_v2_module_update_scenario(
    snippetcompiler_clean,
    modules_v2_dir: str,
    tmpdir: py.path.local,
    local_module_package_index: str,
    package_name_extra: str,
    do_project_update: bool,
) -> None:
    """
    Verify that module installation works correctly when a new version of a module defines a dependency with an extra.

    Dependency tree:

        myv2mod
           --> depmod (V2)
           [--> <extra>]
    """
    index: PipIndex = PipIndex(artifact_dir=str(tmpdir.join(".index")))

    # create module with optional dependency
    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("depmod")),
        new_name="depmod",
        new_requirements=[],
        new_extras={
            "myfeature": [Requirement.parse(package_name_extra)],
        },
        publish_index=index,
    )
    package_without_extra: Requirement = InmantaModuleRequirement.parse("depmod").get_python_package_requirement()
    package_with_extra: Requirement = InmantaModuleRequirement.parse("depmod[myfeature]").get_python_package_requirement()
    package_name: str = str(package_without_extra)

    def assert_installed(*, module_installed: bool = True, extra_installed: bool) -> None:
        installed: abc.Mapping[str, Version] = process_env.get_installed_packages()
        assert (package_name in installed) == module_installed
        assert (package_name_extra in installed) == extra_installed

    # Create myv2mod
    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("myv2mod")),
        new_name="myv2mod",
        new_version=Version("1.0.0"),
        new_content_init_cf="import depmod",
        new_requirements=[package_without_extra],
        publish_index=index,
    )
    # project with dependency on myv2mod without extra
    snippetcompiler_clean.setup_for_snippet(
        "import myv2mod",
        install_project=True,
        index_url=index.url,
        extra_index_url=[local_module_package_index, "https://pypi.org/simple"],
        python_requires=[Requirement.parse("inmanta-module-myv2mod==1.0.0")],
        autostd=False,
    )
    assert_installed(extra_installed=False)

    # Create myv2mod with extra
    shutil.rmtree(tmpdir.join("myv2mod"))
    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("myv2mod")),
        new_name="myv2mod",
        new_version=Version("2.0.0"),
        new_content_init_cf="import depmod",
        new_requirements=[package_with_extra],
        publish_index=index,
    )
    # project with dependency on mymod with extra
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "import myv2mod",
        install_project=not do_project_update,
        index_url=index.url,
        extra_index_url=[local_module_package_index, "https://pypi.org/simple"],
        python_requires=[Requirement.parse("inmanta-module-myv2mod==2.0.0")],
        autostd=False,
    )
    if do_project_update:
        project_tool = ProjectTool()
        project_tool.update(project=project)

    assert_installed(extra_installed=True)


@pytest.mark.slowtest
@pytest.mark.parametrize("package_name_extra", ["inmanta-module-minimalv2module", "lorem"])
@pytest.mark.parametrize("do_project_update", [True, False])
def test_module_install_extra_on_dep_of_v1_module_update_scenario(
    tmpdir,
    snippetcompiler_clean,
    modules_dir: str,
    modules_v2_dir: str,
    local_module_package_index: str,
    package_name_extra: str,
    do_project_update: bool,
) -> None:
    """
    Verify that module installation works correct when a V1 module is updated with a dependency that defines an extra.

    Dependency tree:

        myv1mod
           --> depmod (V2)
           [--> <extra>]
    """
    index: PipIndex = PipIndex(artifact_dir=str(tmpdir.join(".index")))

    # Publish dependency of V1 module (depmod) to python package repo
    package_without_extra: Requirement = InmantaModuleRequirement.parse("depmod").get_python_package_requirement()
    package_with_extra: Requirement = InmantaModuleRequirement.parse("depmod[myfeature]").get_python_package_requirement()
    package_name: str = str(package_without_extra)

    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("depmod")),
        new_name="depmod",
        new_requirements=[],
        new_extras={
            "myfeature": [Requirement.parse(package_name_extra)],
        },
        publish_index=index,
    )

    def assert_installed(*, module_installed: bool = True, extra_installed: bool) -> None:
        installed: abc.Mapping[str, Version] = process_env.get_installed_packages()
        assert (package_name in installed) == module_installed
        assert (package_name_extra in installed) == extra_installed

    # v1 module with dependency on myv1mod without extra
    path_myv1mod = str(tmpdir.join("myv1mod"))
    v1_module_from_template(
        os.path.join(modules_dir, "minimalv1module"),
        path_myv1mod,
        new_name="myv1mod",
        new_requirements=[package_without_extra],
    )
    snippetcompiler_clean.setup_for_snippet(
        "import myv1mod",
        install_project=True,
        add_to_module_path=[str(tmpdir)],
        index_url=index.url,
        extra_index_url=[local_module_package_index, "https://pypi.org/simple"],
        autostd=False,
    )
    assert_installed(extra_installed=False)

    # v1 module with dependency on myv1mod with extra
    shutil.rmtree(path_myv1mod)
    v1_module_from_template(
        os.path.join(modules_dir, "minimalv1module"),
        path_myv1mod,
        new_name="myv1mod",
        new_requirements=[package_with_extra],
    )
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "import myv1mod",
        install_project=not do_project_update,
        add_to_module_path=[str(tmpdir)],
        index_url=index.url,
        extra_index_url=[local_module_package_index, "https://pypi.org/simple"],
        autostd=False,
    )
    if do_project_update:
        project_tool = ProjectTool()
        project_tool.update(project=project)

    assert_installed(extra_installed=True)


@pytest.mark.slowtest
async def test_v1_module_depends_on_third_party_dep_with_extra(
    tmpdir, snippetcompiler_clean, modules_dir: str, index_with_pkgs_containing_optional_deps: str
) -> None:
    """
    Test whether extras on third party python dependencies are correctly handled on V1 modules.
    """
    # Add dependency pkg[optional-a] to module
    v1_module_from_template(
        os.path.join(modules_dir, "minimalv1module"),
        os.path.join(tmpdir, "myv1mod"),
        new_name="myv1mod",
        new_content_init_cf="",
        new_requirements=[Requirement.parse("pkg[optional-a]")],
    )
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "import myv1mod",
        install_project=True,
        add_to_module_path=[str(tmpdir)],
        index_url=index_with_pkgs_containing_optional_deps,
        autostd=False,
    )
    assert project.virtualenv.are_installed(["pkg", "dep-a"])
    assert not project.virtualenv.are_installed(["dep-b"])
    assert not project.virtualenv.are_installed(["dep-c"])

    # Add dependency pkg[optional-a,optional-b] to module
    shutil.rmtree(os.path.join(tmpdir, "myv1mod"))
    v1_module_from_template(
        os.path.join(modules_dir, "minimalv1module"),
        os.path.join(tmpdir, "myv1mod"),
        new_name="myv1mod",
        new_content_init_cf="",
        new_requirements=[Requirement.parse("pkg[optional-a,optional-b]")],
    )
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "import myv1mod",
        install_project=True,
        add_to_module_path=[str(tmpdir)],
        index_url=index_with_pkgs_containing_optional_deps,
        autostd=False,
    )
    assert project.virtualenv.are_installed(["pkg", "dep-a", "dep-b", "dep-c"])


@pytest.mark.slowtest
async def test_v2_module_depends_on_third_party_dep_with_extra(
    tmpdir, snippetcompiler_clean, modules_v2_dir: str, index_with_pkgs_containing_optional_deps: str
) -> None:
    """
    Test whether extras on third party python dependencies are correctly handled on V2 modules.
    """
    index: PipIndex = PipIndex(artifact_dir=str(tmpdir.join(".index")))

    # Add dependency pkg[optional-a] to module
    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("myv2mod")),
        new_name="myv2mod",
        new_version=Version("1.0.0"),
        new_requirements=[Requirement.parse("pkg[optional-a]")],
        publish_index=index,
    )
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "import myv2mod",
        install_project=True,
        python_requires=[Requirement.parse("inmanta-module-myv2mod==1.0.0")],
        index_url=index.url,
        extra_index_url=[index_with_pkgs_containing_optional_deps],
        autostd=False,
    )
    assert project.virtualenv.are_installed(["pkg", "dep-a"])
    assert not project.virtualenv.are_installed(["dep-b"])
    assert not project.virtualenv.are_installed(["dep-c"])

    # Add dependency pkg[optional-a,optional-b] to module
    shutil.rmtree(tmpdir.join("myv2mod"))
    module_from_template(
        os.path.join(modules_v2_dir, "minimalv2module"),
        str(tmpdir.join("myv2mod")),
        new_name="myv2mod",
        new_version=Version("2.0.0"),
        new_requirements=[Requirement.parse("pkg[optional-a,optional-b]")],
        publish_index=index,
    )
    project: Project = snippetcompiler_clean.setup_for_snippet(
        "import myv2mod",
        install_project=True,
        python_requires=[Requirement.parse("inmanta-module-myv2mod==2.0.0")],
        index_url=index.url,
        extra_index_url=[index_with_pkgs_containing_optional_deps],
        autostd=False,
    )
    assert project.virtualenv.are_installed(["pkg", "dep-a", "dep-b", "dep-c"])


async def test_loading_source_and_bytecode(
    server,
    environment,
    snippetcompiler,
    modules_dir: str,
    modules_v2_dir: str,
    tmpdir: py.path.local,
) -> None:
    """The goal of this test is to verify that the exporter does not load both python and pyc files"""
    python_code = """
from inmanta.resources import Resource, resource

@resource("modulev1::Test", agent="agent", id_attribute="name")
class Test(Resource):
    fields = ("name", "agent")
    """

    model_code = """
    entity Test:
        string name
        string agent
    end

    implement Test using std::none
    """

    module_name: str = "minimalv1module"
    module_path: str = str(tmpdir.join("modulev1"))
    v1_module_from_template(
        os.path.join(modules_dir, module_name),
        module_path,
        new_name="modulev1",
        new_content_init_cf=model_code,
        new_content_init_py=python_code,
    )

    plugins_dir: str = os.path.join(module_path, "plugins")
    init_py_file: str = os.path.join(plugins_dir, "__init__.py")
    py_compile.compile(file=init_py_file, cfile=init_py_file + "c", doraise=True)

    snippetcompiler.setup_for_snippet(
        "import modulev1\nmodulev1::Test(name='abc', agent='def')",
        add_to_module_path=[str(tmpdir)],
    )
    await snippetcompiler.do_export_and_deploy(do_raise=False)

    code_manager = loader.CodeManager()
    for type_name, resource_definition in resources.resource.get_resources():
        code_manager.register_code(type_name, resource_definition)

    module_code = False
    for name, source_info in code_manager.get_types():
        for info in source_info:
            if info.module_name == "inmanta_plugins.modulev1":
                module_code = True
                assert info.path[-4:] == ".pyc"

    assert module_code


@pytest.mark.skipif(
    "inmanta-core" in process_env.get_installed_packages(only_editable=True),
    reason="Inmanta package protection in env.install_* not compatible with editable core in non-inherited venv.",
)
@pytest.mark.slowtest
async def test_v2_module_editable_with_links(tmpvenv_active: tuple[py.path.local, py.path.local], modules_v2_dir: str) -> None:
    """
    One possible implementation mechanism for editable installs
    (https://setuptools.pypa.io/en/latest/userguide/development_mode.html#how-editable-installations-work) is to use a farm of
    symlinks to the actual source files. Since setuptools is not necessarily aware of a module's non-Python files we need to
    ensure our module discovery implementation is robust against this. Because setuptools provides no guarantees as to which
    mechanism is used under which circumstances, we mimic such an editable install here.
    """
    module_dir: str = os.path.join(modules_v2_dir, "minimalv2module")
    # start with non-editable install to populate site-packages with appropriate metadata (editable install would create pth
    # files if a different mechanism is picked so we want to avoid that).
    process_env.install_for_config(
        requirements=[], paths=[LocalPackagePath(path=module_dir, editable=False)], config=PipConfig(use_system_config=True)
    )
    # replace module dir in site-packages with symlink
    rel_path_src: str = os.path.join(const.PLUGINS_PACKAGE, "minimalv2module")
    module_dir_site_packages: str = os.path.join(process_env.site_packages_dir, rel_path_src)
    shutil.rmtree(module_dir_site_packages)
    os.symlink(os.path.join(module_dir, rel_path_src), module_dir_site_packages)

    # verify that module can be found
    module: Optional[ModuleV2] = ModuleV2Source().get_installed_module(DummyProject(autostd=False), "minimalv2module")
    assert module is not None
    assert module.path == module_dir


def test_cross_module_dependency(local_module_package_index: str, snippetcompiler, capsys) -> None:
    """
    This test checks that the python code living in the inmanta_plugins dir of a module ('anothermod' in this test case)
    that is not loaded, can be used from the plugins of another module. ('cross_module_dependency' in this test case)
    """

    def check_name_space(name_space: dict[str, object], includes: Sequence[str], excludes: Sequence[str]) -> None:
        """
        Check that all the items in `includes` are present in `name_space`
        and that no item in `excludes` is present in `name_space`
        """
        for x in includes:
            assert x in name_space
        for x in excludes:
            assert x not in name_space

    project: Project = snippetcompiler.setup_for_snippet(
        """
import cross_module_dependency

cross_module_dependency::print_message('message from project model')
cross_module_dependency::call_to_triple_from_another_mod('triple this string')

        """.strip(),
        index_url=local_module_package_index,
        python_requires=[
            InmantaModuleRequirement.parse("minimalv2module").get_python_package_requirement(),
            InmantaModuleRequirement.parse("cross_module_dependency").get_python_package_requirement(),
        ],
        autostd=False,
    )

    check_name_space(
        name_space=project.modules, includes=["cross_module_dependency"], excludes=["minimalv2module", "anothermod"]
    )
    check_name_space(
        name_space=sys.modules,
        includes=[],
        excludes=["inmanta_plugins.cross_module_dependency", "inmanta_plugins.minimalv2module", "inmanta_plugins.anothermod"],
    )

    types, _ = compiler.do_compile()

    out, _ = capsys.readouterr()
    output = out.strip()

    expected_output: list[str] = [
        "message from project model",
        "triple this string" * 3,
        "message from cross_module_dependency model",
    ]
    for line in expected_output:
        assert line in output

    check_name_space(
        name_space=project.modules, includes=["cross_module_dependency"], excludes=["minimalv2module", "anothermod"]
    )
    check_name_space(
        name_space=sys.modules,
        includes=["inmanta_plugins.cross_module_dependency", "inmanta_plugins.anothermod"],
        excludes=["inmanta_plugins.minimalv2module"],
    )
    # Check that the flag_plugin Plugin object is kept in PluginMeta
    check_name_space(
        name_space=plugins.PluginMeta.get_functions(),
        includes=[
            "cross_module_dependency::print_message",
            "cross_module_dependency::call_to_triple_from_another_mod",
            "anothermod::flag_plugin",
        ],
        excludes=[],
    )
    # ...but that the corresponding statement is NOT created
    check_name_space(
        name_space=types,
        includes=[
            "cross_module_dependency::print_message",
            "cross_module_dependency::call_to_triple_from_another_mod",
        ],
        excludes=["anothermod::flag_plugin"],
    )


def test_cross_module_dependency_sub_module(local_module_package_index: str, snippetcompiler, capsys) -> None:
    """
    This test checks that the plugins of a top-level module still work even if one of its sub-modules is not loaded.
    """

    snippetcompiler.setup_for_snippet(
        """
import complex_module_dependencies_mod2

complex_module_dependencies_mod2::cmd_mod2()
        """.strip(),
        index_url=local_module_package_index,
        python_requires=[
            InmantaModuleRequirement.parse("complex_module_dependencies_mod2").get_python_package_requirement(),
        ],
        autostd=False,
    )

    compiler.do_compile()

    out, _ = capsys.readouterr()
    output = out.strip()

    expected_output: str = "Hello from complex_module_dependencies_mod2"

    assert expected_output in output



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import argparse
from collections import abc
from typing import Callable, Optional

FunctionType = Callable[[argparse.Namespace], None]
ParserConfigType = Callable[[argparse.ArgumentParser, abc.Sequence[argparse.ArgumentParser]], None]


class CLIException(Exception):
    def __init__(self, *args: str, exitcode: int) -> None:
        self.exitcode = exitcode
        super().__init__(*args)


class ShowUsageException(Exception):
    """
    Raise this exception to show the usage message of the given level
    """


class Commander:
    """
    This class handles commands
    """

    __command_functions: dict[str, dict[str, object]] = {}

    @classmethod
    def add(
        cls,
        name: str,
        function: FunctionType,
        help_msg: str,
        parser_config: Optional[ParserConfigType],
        require_project: bool = False,
        aliases: list[str] = [],
        add_verbose_flag: bool = True,
    ) -> None:
        """
        Add a new export function
        """
        if name in cls.__command_functions:
            raise Exception("Command %s already registered" % name)

        cls.__command_functions[name] = {
            "function": function,
            "help": help_msg,
            "parser_config": parser_config,
            "require_project": require_project,
            "aliases": aliases,
            "add_verbose_flag": add_verbose_flag,
        }

    config = None

    @classmethod
    def reset(cls) -> None:
        """
        Resets the list of commands
        """
        cls.__command_functions = {}

    @classmethod
    def commands(cls) -> dict[str, dict[str, object]]:
        """
        Return a list of commands
        """
        return cls.__command_functions


class command:  # noqa: N801
    """
    A decorator that registers an export function

    :param add_verbose_flag: Set this to false to prevent automatically adding the verbose option to the registered command.
    """

    def __init__(
        self,
        name: str,
        help_msg: str,
        parser_config: Optional[ParserConfigType] = None,
        require_project: bool = False,
        aliases: list[str] = [],
        add_verbose_flag: bool = True,
    ) -> None:
        self.name = name
        self.help = help_msg
        self.require_project = require_project
        self.parser_config = parser_config
        self.aliases = aliases
        self.add_verbose_flag = add_verbose_flag

    def __call__(self, function: FunctionType) -> FunctionType:
        """
        The wrapping
        """
        Commander.add(
            self.name, function, self.help, self.parser_config, self.require_project, self.aliases, self.add_verbose_flag
        )
        return function
